import os
from io import BytesIO

import PIL as pillow

from pdfmajor.BMPWriter import BMPWriter
from pdfmajor.utils import Bbox
from pdfmajor.parser.PDFStream import PDFStream, LITERALS_DCT_DECODE
from pdfmajor.parser.constants import LIT

from ._base import LTComponent

LITERAL_DEVICE_GRAY = LIT('DeviceGray')
LITERAL_DEVICE_RGB = LIT('DeviceRGB')
LITERAL_DEVICE_CMYK = LIT('DeviceCMYK')

##  LTImage
##
class LTImage(LTComponent):

    def __init__(self, name: str, stream: PDFStream, bbox: Bbox):
        LTComponent.__init__(self, bbox)
        self.name = name
        self.stream = stream
        self.srcsize = (stream.get_any(('W', 'Width')),
                        stream.get_any(('H', 'Height')))
        self.imagemask = stream.get_any(('IM', 'ImageMask'))
        self.bits = stream.get_any(('BPC', 'BitsPerComponent'), 1)
        self.colorspace = stream.get_any(('CS', 'ColorSpace'))
        if not isinstance(self.colorspace, list):
            self.colorspace = [self.colorspace]
        return

    def __repr__(self):
        return ('<%s(%s) %s %r>' %
                (self.__class__.__name__, self.name,
                 str(self.bbox), self.srcsize))
    
    def save_image(self, outdir: str, file_name: str = None):
        stream = self.stream
        filters = stream.get_filters()
        width = self.bbox.width
        height = self.bbox.height
        bits = self.bits

        if len(filters) == 1 and filters[0][0] in LITERALS_DCT_DECODE:
            ext = '.jpg'
        elif (bits == 1 or
                bits == 8 and self.colorspace in (LITERAL_DEVICE_RGB, LITERAL_DEVICE_GRAY)):
            ext = '.w%dh%d.bmp' % (width, height)
        else:
            ext = '.b%dw%dh%d.img' % (bits, width, height)

        if file_name is None:
            file_name = self.name + ext
        else:
            file_name += ext
        path = os.path.join(outdir, file_name)
        
        with open(path, 'wb') as fp:
            if ext == '.jpg':
                raw_data = stream.get_rawdata()
                if LITERAL_DEVICE_CMYK in self.colorspace:
                    ifp = BytesIO(raw_data)
                    i = pillow.Image.open(ifp)
                    i = pillow.ImageChops.invert(i)
                    i = i.convert('RGB')
                    i.save(fp, 'JPEG')
                else:
                    fp.write(raw_data)
            elif bits == 1:
                BMPWriter.write_all(fp, stream, 1, width, height, (width+7)//8)
            elif bits == 8 and self.colorspace is LITERAL_DEVICE_RGB:
                BMPWriter.write_all(fp, stream, 24, width, height, width*3)
            elif bits == 8 and self.colorspace is LITERAL_DEVICE_GRAY:
                BMPWriter.write_all(fp, stream, 8, width, height, width)
            else:
                fp.write(stream.get_data())
        return path



from typing import List, Tuple

from pdfmajor.utils import Bbox
from ..PDFGraphicState import PDFColor
from ..Curves import CurvePath

from ._base import LTComponent, Bbox

class LTCurve(LTComponent):

    def __init__(self, 
        linewidth: float, 
        paths: List[CurvePath], 
        bbox: Bbox,
        stroke: PDFColor, 
        fill: PDFColor, 
        evenodd: bool, 
    ):
        LTComponent.__init__(self, bbox)
        self.paths = paths
        self.linewidth = linewidth
        self.evenodd = evenodd
        self.stroke = stroke
        self.fill = fill

##  LTLine
##
class LTLine(LTCurve):
    pass

##  LTHorizontalLine
##
class LTHorizontalLine(LTLine):
    pass

##  LTHorizontalLine
##
class LTVerticalLine(LTLine):
    pass

##  LTRect
##
class LTRect(LTCurve):
    pass




import struct
from io import BytesIO
                
from ..parser.PDFStream import LITERALS_DCT_DECODE, PDFStream

def align32(x):
    return ((x+3)//4)*4

##  BMPWriter
##
class BMPWriter(object):

    @classmethod
    def write_all(cls, fp, stream: PDFStream, bits: int, width: float, height: float, step: float):
        bmp = cls(fp, 1, width, height)
        data = stream.get_data()
        i = 0
        for y in range(height):
            bmp.write_line(y, data[i:i+step])
            i += step

    def __init__(self, fp, bits, width, height):
        self.fp = fp
        self.bits = bits
        self.width = width
        self.height = height
        if bits == 1:
            ncols = 2
        elif bits == 8:
            ncols = 256
        elif bits == 24:
            ncols = 0
        else:
            raise ValueError(bits)
        self.linesize = align32((self.width*self.bits+7)//8)
        self.datasize = self.linesize * self.height
        headersize = 14+40+ncols*4
        info = struct.pack('<IiiHHIIIIII', 40, self.width, self.height, 1, self.bits, 0, self.datasize, 0, 0, ncols, 0)
        assert len(info) == 40, str(len(info))
        header = struct.pack('<ccIHHI', b'B', b'M', headersize+self.datasize, 0, 0, headersize)
        assert len(header) == 14, str(len(header))
        self.fp.write(header)
        self.fp.write(info)
        if ncols == 2:
            # B&W color table
            for i in (0, 255):
                self.fp.write(struct.pack('BBBx', i, i, i))
        elif ncols == 256:
            # grayscale color table
            for i in range(256):
                self.fp.write(struct.pack('BBBx', i, i, i))
        self.pos0 = self.fp.tell()
        self.pos1 = self.pos0 + self.datasize
        return

    def write_line(self, y, data):
        self.fp.seek(self.pos1 - (y+1)*self.linesize)
        self.fp.write(data)
        return



from typing import List

from pdfmajor.utils import apply_matrix_norm, apply_matrix_pt, matrix2str, Bbox, Point

from ..PDFGraphicState import PDFColor
from ..PDFTextState import PDFFont
from ..PDFTextState import PDFTextState

from ._base import LTContainer
from .LTCharBlock import LTCharBlock
from .utils.textdecoder import decode_text_seq


##  LTTextBlock
##
class LTTextBlock(LTContainer):
    def __init__(self, x0=None, y0=None, x1=None, y1=None):
        LTContainer.__init__(self, 
            Bbox(x0, y0, x1, y1),
            []
        )
    
    def add_char_block(self, seq: bytearray, ctm: tuple, textstate: PDFTextState, color: PDFColor):
        (char_meta_datas, textstate, color) = decode_text_seq(seq, ctm, textstate, color)
        char_block = LTCharBlock(
            chars=char_meta_datas,
            color=color,
            textstate=textstate
        )
        if None in [self.bbox.x0, self.bbox.x1, self.bbox.y0, self.bbox.y1]:
            self.bbox.x0 = char_block.x0
            self.bbox.x1 = char_block.x1
            self.bbox.y0 = char_block.y0
            self.bbox.y1 = char_block.y1
        else:
            self.bbox.x0 = min([self.bbox.x0, char_block.x0])
            self.bbox.x1 = max([self.bbox.x1, char_block.x1])
            self.bbox.y0 = min([self.bbox.y0, char_block.y0])
            self.bbox.y1 = max([self.bbox.y1, char_block.y1])
        self.add(char_block)


import io
import os
import re

from typing import List

from .html import convert_to_html
from .xml import convert_to_xml
from .json import convert_to_json
from .text import convert_to_text
from ..utils import logging

def convert_file(
        input_file: str, 
        output_file: str = None, 
        image_folder_path: str = None,
        codec: str = 'utf-8',
        maxpages: int = 0, 
        password: str = None, 
        caching: bool = True, 
        check_extractable: bool = True,
        pagenos: List[int] = None,
        dont_export_images: bool = False,
        debug_level: int = logging.WARNING,
        out_type: str = 'html',
    ):
    if output_file is None:
        output_file = os.path.join(
            os.path.dirname(input_file.name),
            os.path.basename(
                re.sub(r"\.\w+$", "."+out_type, input_file.name)
            )
        )
    if re.search(re.escape(out_type) + r'$', output_file ) is None:
        raise Exception("Please make sure that the file name and output type match!")
    if out_type == 'html':
        return convert_to_html(
            input_file_path=input_file, 
            output_file_path=output_file, 
            image_folder_path=image_folder_path,
            dont_export_images=dont_export_images,
            codec=codec,
            maxpages=maxpages, 
            password=password, 
            caching=caching, 
            check_extractable=check_extractable,
            pagenos=pagenos,
            debug_level=debug_level,
        )
    elif out_type == 'xml':
        return convert_to_xml(
            input_file_path=input_file, 
            output_file_path=output_file, 
            image_folder_path=image_folder_path,
            codec=codec,
            maxpages=maxpages, 
            password=password, 
            caching=caching, 
            check_extractable=check_extractable,
            pagenos=pagenos,
            dont_export_images=dont_export_images,
            debug_level=debug_level,
        )
    elif out_type == 'json':
        return convert_to_json(
            input_file_path=input_file, 
            output_file_path=output_file, 
            image_folder_path=image_folder_path,
            codec=codec,
            maxpages=maxpages, 
            password=password, 
            caching=caching, 
            check_extractable=check_extractable,
            pagenos=pagenos,
            dont_export_images=dont_export_images,
            debug_level=debug_level,
        )
    elif out_type == 'text':
        return convert_to_text(
            input_file_path=input_file, 
            output_file_path=output_file, 
            image_folder_path=image_folder_path,
            codec=codec,
            maxpages=maxpages, 
            password=password, 
            caching=caching, 
            check_extractable=check_extractable,
            pagenos=pagenos,
            dont_export_images=dont_export_images,
            debug_level=debug_level,
        )
    else: raise Exception("Please specify out_type as 'html' or 'xml' or 'json' or 'text'")


from typing import List

from ._base import LTContainer
from .LTChar import LTChar

from pdfmajor.utils import apply_matrix_norm, apply_matrix_pt, matrix2str, Bbox, Point
from ..PDFGraphicState import PDFColor
from ..PDFTextState import PDFFont
from ..PDFTextState import PDFTextState

##  LTCharBlock
##
class LTCharBlock(LTContainer):

    def __init__(self, 
        chars: list, # [str, List[Point]]
        textstate: PDFTextState, 
        color: PDFColor
    ):
        c, ((x0, y0), (x1, y1)) = chars[0]
        ltchars = [
            LTChar(
                bbox=Bbox(x0, y0, x1, y1),
                char= c,
                textstate=textstate, 
                color=color
            )
        ]

        for c, pts in chars[1:]:
            ((nx0, ny0), (nx1, ny1)) = pts
            x0 = min([ x0, x1, nx0, nx1 ])
            x1 = max([ x0, x1, nx0, nx1 ])
            y1 = max([ y0, y1, ny0, ny1 ])
            y0 = min([ y0, y1, ny0, ny1 ])
            ltchars.append(LTChar(
                bbox=Bbox(nx0, ny0, nx1, ny1),
                char= c,
                textstate=textstate, 
                color=color
            ))

        LTContainer.__init__(self, Bbox(x0, y0, x1, y1), ltchars)
        
        self.textstate = textstate
        self.color = color

        if self.textstate.font.is_vertical():
            self.size = self.width
        else:
            self.size = self.height
        return

    def __repr__(self):
        return f"""<{self.__class__.__name__} font="{self.fontname}" text="{self.get_text()}"/>"""

    def get_text(self):
        return "".join([ c.get_text() for c in self ])

    @property
    def font(self):
        return self.textstate.font

    @property
    def fontname(self) -> str:
        return self.textstate.font.fontname


# PDFFont
from abc import abstractmethod
from io import BytesIO

from pdfmajor.parser.EncodingDB import EncodingDB
from pdfmajor.parser.PSStackParser import LIT
from pdfmajor.parser.PSStackParser import PSLiteral
from pdfmajor.parser.PSStackParser import literal_name
from pdfmajor.parser.PDFStream import resolve1
from pdfmajor.parser.PDFStream import int_value
from pdfmajor.parser.PDFStream import num_value
from pdfmajor.parser.PDFStream import list_value
from pdfmajor.parser.PDFStream import PDFException
from pdfmajor.parser.PDFStream import PDFStream
from pdfmajor.parser.cmapdb import FileUnicodeMap
from pdfmajor.parser.cmapdb import CMapParser

##  Literals
##
LITERAL_STANDARD_ENCODING = LIT('StandardEncoding')
LITERAL_TYPE1C = LIT('Type1C')


class PDFFont(object):
    class PDFFontError(PDFException): pass
    class PDFUnicodeNotDefined(PDFFontError): pass

    def __init__(self, descriptor, widths, default_width=None):
        self.descriptor = descriptor
        self.widths = widths
        self.fontname = resolve1(descriptor.get('FontName', 'unknown'))
        if isinstance(self.fontname, PSLiteral):
            self.fontname = literal_name(self.fontname)
        self.flags = int_value(descriptor.get('Flags', 0))
        self.ascent = num_value(descriptor.get('Ascent', 0))
        self.descent = num_value(descriptor.get('Descent', 0))
        self.italic_angle = num_value(descriptor.get('ItalicAngle', 0))
        self.default_width = default_width or num_value(descriptor.get('MissingWidth', 0))
        self.leading = num_value(descriptor.get('Leading', 0))
        self.bbox = list_value(descriptor.get('FontBBox', (0, 0, 0, 0)))
        self.hscale = self.vscale = .001
        return

    def __repr__(self):
        return '<PDFFont>'

    def is_vertical(self):
        return False

    def is_multibyte(self):
        return False

    def decode(self, bytes):
        return bytearray(bytes)  # map(ord, bytes)

    def get_ascent(self):
        return self.ascent * self.vscale

    def get_descent(self):
        return self.descent * self.vscale

    def get_width(self):
        w = self.bbox[2]-self.bbox[0]
        if w == 0:
            w = -self.default_width
        return w * self.hscale

    def get_height(self):
        h = self.bbox[3]-self.bbox[1]
        if h == 0:
            h = self.ascent - self.descent
        return h * self.vscale

    def char_width(self, cid):
        try:
            return self.widths[cid] * self.hscale
        except KeyError:
            try:
                return self.widths[self.to_unichr(cid)] * self.hscale
            except (KeyError, self.PDFUnicodeNotDefined):
                return self.default_width * self.hscale

    def char_disp(self, cid):
        return 0

    def string_width(self, s):
        return sum(self.char_width(cid) for cid in self.decode(s))

    @abstractmethod
    def to_unichr(self, cid):
        raise NotImplementedError
    
    @property
    def font_weight(self):
        if 'FontWeight' in self.descriptor.keys():
            return self.descriptor['FontWeight']
        elif 'bold' in self.fontname.lower():
            return 'bold'
        else:
            return None
    
    @property
    def is_bold(self):
        return self.font_weight == 'bold' or self.font_weight > 549

# PDFSimpleFont
class PDFSimpleFont(PDFFont):

    def __init__(self, descriptor, widths, spec):
        # Font encoding is specified either by a name of
        # built-in encoding or a dictionary that describes
        # the differences.
        if 'Encoding' in spec:
            encoding = resolve1(spec['Encoding'])
        else:
            encoding = LITERAL_STANDARD_ENCODING
        if isinstance(encoding, dict):
            name = literal_name(encoding.get('BaseEncoding', LITERAL_STANDARD_ENCODING))
            diff = list_value(encoding.get('Differences', []))
            self.cid2unicode = EncodingDB.get_encoding(name, diff)
        else:
            self.cid2unicode = EncodingDB.get_encoding(literal_name(encoding))
        self.unicode_map = None
        if 'ToUnicode' in spec:
            strm = PDFStream.validated_stream(spec['ToUnicode'])
            self.unicode_map = FileUnicodeMap()
            CMapParser(self.unicode_map, BytesIO(strm.get_data())).run()
        PDFFont.__init__(self, descriptor, widths)
        return

    def to_unichr(self, cid):
        if self.unicode_map:
            try:
                return self.unicode_map.get_unichr(cid)
            except KeyError:
                pass
        try:
            return self.cid2unicode[cid]
        except KeyError:
            raise self.PDFUnicodeNotDefined(None, cid)



from pdfmajor.utils import mult_matrix, isnumber, apply_matrix_pt, translate_matrix

from ..PDFTextState import PDFTextState
from ..PDFGraphicState import PDFGraphicState, PDFColorSpace
from .PDFItem import PDFText
from .Char import Char, Bbox

class TextDecoder:

    @classmethod
    def make_pdftext(cls, seq: bytearray, ctm, textstate: PDFTextState, graphicstate: PDFGraphicState) -> PDFText:
        textstate.matrix = mult_matrix(textstate.matrix, ctm)
        textstate.scaling = textstate.scaling * .01
        textstate.charspace = textstate.charspace * textstate.scaling
        textstate.wordspace = textstate.wordspace * textstate.scaling
        ncs = graphicstate.ncolor.color_space
       
        if textstate.font.is_multibyte():
            textstate.wordspace = 0
        dxscale = .001 * textstate.fontsize * textstate.scaling

        if textstate.font.is_vertical():
            return cls.__render_string_along(1,
                seq, textstate, dxscale, ncs, graphicstate
            )
        else:
            return cls.__render_string_along(0,
                seq, textstate, dxscale, ncs, graphicstate
            )

    @classmethod
    def __render_string_along(cls, idx: int, seq: bytearray, 
        textstate: PDFTextState, dxscale: float, ncs: PDFColorSpace, graphicstate: PDFGraphicState) -> PDFText:
        needcharspace = False
        char_meta_datas = []
        for obj in seq:
            if isnumber(obj):
                textstate.linematrix[idx] -= obj*dxscale
                needcharspace = True
            else:
                for cid in textstate.font.decode(obj):
                    if needcharspace:
                        textstate.linematrix[idx] += textstate.charspace
                    
                    text = textstate.font.to_unichr(cid)
                    assert isinstance(text, str), str(type(text))

                    matrix = translate_matrix(textstate.matrix, textstate.linematrix)
                    adv, bbox = cls.__compute_char_bbox(matrix, cid, textstate)
                    textstate.linematrix[idx] += adv

                    char_meta_datas.append(
                        Char( text, Bbox.from_points(bbox), )
                    )
                    if cid == 32 and textstate.wordspace:
                        textstate.linematrix[idx] += textstate.wordspace
                    needcharspace = True
                    
        return PDFText(char_meta_datas, textstate, graphicstate)

    @classmethod
    def __compute_char_bbox(cls, matrix, char_id: int, textstate: PDFTextState):
        font = textstate.font
        fontsize = textstate.fontsize
        rise = textstate.rise
        scaling = textstate.scaling

        textwidth = font.char_width(char_id)
        textdisp = font.char_disp(char_id)
        adv = textwidth * fontsize * scaling

        # compute the boundary rectangle.
        if font.is_vertical():
            # vertical
            width = font.get_width() * fontsize
            (vx, vy) = textdisp
            if vx is None:
                vx = width * 0.5
            else:
                vx = vx * fontsize * .001
            vy = (1000 - vy) * fontsize * .001
            tx = -vx
            ty = vy + rise
            bll = (tx, ty+adv)
            bur = (tx+width, ty)
        else:
            # horizontal
            height = font.get_height() * fontsize
            descent = font.get_descent() * fontsize
            ty = descent + rise
            bll = (0, ty)
            bur = (adv, ty+height)
        (x0, y0) = apply_matrix_pt(matrix, bll)
        (x1, y1) = apply_matrix_pt(matrix, bur)
        if x1 < x0:
            (x0, x1) = (x1, x0)
        if y1 < y0:
            (y0, y1) = (y1, y0)
        
        return (adv, ((x0, y0), (x1, y1)))


class settings:
    STRICT = False 

from .encoding import *

from .ascii85 import *
from .fontmetrics import FONT_METRICS
from .generic import * 
from .lzw import * 
from .matrix import *
from .text import *
from .types import *
from .log import *


from ._base import LTComponent
from pdfmajor.utils import apply_matrix_norm, apply_matrix_pt, matrix2str, Bbox
from ..PDFGraphicState import PDFColor
from ..PDFTextState import PDFTextState, PDFFont

##  LTChar
##
class LTChar(LTComponent):

    def __init__(self, 
        bbox: Bbox, 
        char: str, 
        textstate: PDFTextState, 
        color: PDFColor
    ):
        self._text = char
        self.textstate = textstate
        self.color = color
        LTComponent.__init__(self, bbox)
        if self.textstate.font.is_vertical():
            self.size = self.width
        else:
            self.size = self.height
        return

    def __repr__(self):
        return (
            '<%s %s font=%r text=%r>' %
            (
                self.__class__.__name__, 
                str(self.bbox),
                self.fontname, 
                self.get_text()
            )
        )

    def get_text(self):
        return self._text

    @property
    def font(self):
        return self.textstate.font

    @property
    def fontname(self) -> str:
        return self.font.fontname


import re
import operator

from typing import List
from io import BytesIO

from ..utils import int2byte, choplist, settings

from .PSStackParser import literal_name
from .PSStackParser import PSStackParser
from .PSStackParser import PSEOF, PSTypeError
from .PSStackParser import KWD

from .PDFStream import PDFStream

class PDFContentParser(PSStackParser):

    def __init__(self, streams: List[PDFStream]):
        self.streams = streams
        self.istream = 0
        PSStackParser.__init__(self, None)
        return

    def fillfp(self):
        if not self.fp:
            if self.istream < len(self.streams):
                strm = PDFStream.validated_stream(self.streams[self.istream])
                self.istream += 1
            else:
                raise PSEOF('Unexpected EOF, file truncated?')
            self.fp = BytesIO(strm.get_data())
        return

    def seek(self, pos):
        self.fillfp()
        PSStackParser.seek(self, pos)
        return

    def fillbuf(self):
        if self.charpos < len(self.buf):
            return
        while 1:
            self.fillfp()
            self.bufpos = self.fp.tell()
            self.buf = self.fp.read(self.BUFSIZ)
            if self.buf:
                break
            self.fp = None
        self.charpos = 0
        return

    def get_inline_data(self, pos, target=b'EI'):
        self.seek(pos)
        i = 0
        data = b''
        while i <= len(target):
            self.fillbuf()
            if i:
                c = operator.getitem(self.buf,self.charpos)
                c = int2byte(c)
                data += c
                self.charpos += 1
                if len(target) <= i and c.isspace():
                    i += 1
                elif i < len(target) and c == (int2byte(target[i])):
                    i += 1
                else:
                    i = 0
            else:
                try:
                    j = self.buf.index(target[0], self.charpos)
                    data += self.buf[self.charpos:j+1]
                    self.charpos = j+1
                    i = 1
                except ValueError:
                    data += self.buf[self.charpos:]
                    self.charpos = len(self.buf)
        data = data[:-(len(target)+1)]  # strip the last part
        data = re.sub(br'(\x0d\x0a|[\x0d\x0a])$', b'', data)
        return (pos, data)

    def flush(self):
        self.add_results(*self.popall())
        return

    KEYWORD_BI = KWD(b'BI')
    KEYWORD_ID = KWD(b'ID')
    KEYWORD_EI = KWD(b'EI')

    def do_keyword(self, pos, token):
        if token is self.KEYWORD_BI:
            # inline image within a content stream
            self.start_type(pos, 'inline')
        elif token is self.KEYWORD_ID:
            try:
                (_, objs) = self.end_type('inline')
                if len(objs) % 2 != 0:
                    raise PSTypeError('Invalid dictionary construct: %r' % objs)
                d = dict((literal_name(k), v) for (k, v) in choplist(2, objs))
                (pos, data) = self.get_inline_data(pos+len(b'ID '))
                obj = PDFStream(d, data)
                self.push((pos, obj))
                self.push((pos, self.KEYWORD_EI))
            except PSTypeError:
                if settings.STRICT:
                    raise
        else:
            self.push((pos, token))
        return



from pdfmajor.utils import FONT_METRICS, choplist, isnumber


def get_widths(seq):
    widths = {}
    r = []
    for v in seq:
        if isinstance(v, list):
            if r:
                char1 = r[-1]
                for (i, w) in enumerate(v):
                    widths[char1+i] = w
                r = []
        elif isnumber(v):
            r.append(v)
            if len(r) == 3:
                (char1, char2, w) = r
                for i in range(char1, char2+1):
                    widths[i] = w
                r = []
    return widths
assert get_widths([1]) == {}
assert get_widths([1,2,3]) == {1:3, 2:3}
assert get_widths([1,[2,3],6,[7,8]]) == {1:2,2:3, 6:7,7:8}

def get_widths2(seq):
    widths = {}
    r = []
    for v in seq:
        if isinstance(v, list):
            if r:
                char1 = r[-1]
                for (i, (w, vx, vy)) in enumerate(choplist(3, v)):
                    widths[char1+i] = (w, (vx, vy))
                r = []
        elif isnumber(v):
            r.append(v)
            if len(r) == 5:
                (char1, char2, w, vx, vy) = r # pylint: disable=E0632
                for i in range(char1, char2+1):
                    widths[i] = (w, (vx, vy))
                r = []
    return widths
assert get_widths2([1]) == {}
assert get_widths2([1,2,3,4,5]) == {1:(3, (4,5)), 2:(3, (4,5))}
assert get_widths2([1,[2,3,4,5],6,[7,8,9]]) == {1:(2, (3,4)), 6:(7, (8,9))}


##  FontMetricsDB
##
class FontMetricsDB(object):

    @classmethod
    def get_metrics(klass, fontname):
        return FONT_METRICS[fontname]





from pdfmajor.parser.PSStackParser import PSStackParser
from pdfmajor.parser.PSStackParser import PSEOF, KWD
from pdfmajor.parser.PSStackParser import PSLiteral
from pdfmajor.parser.PSStackParser import literal_name
from pdfmajor.parser.EncodingDB import name2unicode

##  Type1FontHeaderParser
##
class Type1FontHeaderParser(PSStackParser):

    KEYWORD_BEGIN = KWD(b'begin')
    KEYWORD_END = KWD(b'end')
    KEYWORD_DEF = KWD(b'def')
    KEYWORD_PUT = KWD(b'put')
    KEYWORD_DICT = KWD(b'dict')
    KEYWORD_ARRAY = KWD(b'array')
    KEYWORD_READONLY = KWD(b'readonly')
    KEYWORD_FOR = KWD(b'for')
    KEYWORD_FOR = KWD(b'for')

    def __init__(self, data):
        PSStackParser.__init__(self, data)
        self._cid2unicode = {}
        return

    def get_encoding(self):
        while 1:
            try:
                (cid, name) = self.nextobject()
            except PSEOF:
                break
            try:
                self._cid2unicode[cid] = name2unicode(name)
            except KeyError:
                pass
        return self._cid2unicode

    def do_keyword(self, pos, token):
        if token is self.KEYWORD_PUT:
            ((_, key), (_, value)) = self.pop(2) # pylint: disable=E0632
            if (isinstance(key, int) and
                isinstance(value, PSLiteral)):
                self.add_results((key, literal_name(value)))
        return




##  PS Exceptions
##
class PSException(Exception):
    pass


class PSEOF(PSException):
    pass


class PSSyntaxError(PSException):
    pass


class PSTypeError(PSException):
    pass


class PSValueError(PSException):
    pass


from .PDFObjRef import PDFObjRef
from ...utils import settings, isnumber, int2byte
from .exceptions import PDFTypeError

from logging import getLogger

log = getLogger(__name__)

# resolve
def resolve1(x, default=None):
    """Resolves an object.

    If this is an array or dictionary, it may still contains
    some indirect objects inside.
    """
    while isinstance(x, PDFObjRef):
        x = x.resolve(default=default)
    return x


def resolve_all(x, default=None):
    """Recursively resolves the given object and all the internals.

    Make sure there is no indirect reference within the nested object.
    This procedure might be slow.
    """
    while isinstance(x, PDFObjRef):
        x = x.resolve(default=default)
    if isinstance(x, list):
        x = [resolve_all(v, default=default) for v in x]
    elif isinstance(x, dict):
        for (k, v) in x.iteritems():
            x[k] = resolve_all(v, default=default)
    return x


def decipher_all(decipher, objid, genno, x):
    """Recursively deciphers the given object.
    """
    if isinstance(x, bytes):
        return decipher(objid, genno, x)
    if isinstance(x, list):
        x = [decipher_all(decipher, objid, genno, v) for v in x]
    elif isinstance(x, dict):
        for (k, v) in x.items():
            x[k] = decipher_all(decipher, objid, genno, v)
    return x
# Type cheking
def int_value(x):
    x = resolve1(x)
    if not isinstance(x, int):
        if settings.STRICT:
            raise PDFTypeError('Integer required: %r' % x)
        return 0
    return x


def float_value(x):
    x = resolve1(x)
    if not isinstance(x, float):
        if settings.STRICT:
            raise PDFTypeError('Float required: %r' % x)
        return 0.0
    return x


def num_value(x):
    x = resolve1(x)
    if not isnumber(x):
        if settings.STRICT:
            raise PDFTypeError('Int or Float required: %r' % x)
        return 0
    return x


def str_value(x):
    x = resolve1(x)
    if not isinstance(x, bytes):
        if settings.STRICT:
            raise PDFTypeError('String required: %r' % x)
        return ''
    return x


def list_value(x):
    x = resolve1(x)
    if not isinstance(x, (list, tuple)):
        if settings.STRICT:
            raise PDFTypeError('List required: %r' % x)
        return []
    return x


def dict_value(x):
    x = resolve1(x)
    if not isinstance(x, dict):
        if settings.STRICT:
            log.error('PDFTypeError : Dict required: %r', x)
            raise PDFTypeError('Dict required: %r' % x)
        return {}
    return x
    
#
# RunLength decoder (Adobe version) implementation based on PDF Reference
# version 1.4 section 3.3.4.
#
#  * public domain *
#
def rldecode(data):
    """
    RunLength decoder (Adobe version) implementation based on PDF Reference
    version 1.4 section 3.3.4:
        The RunLengthDecode filter decodes data that has been encoded in a
        simple byte-oriented format based on run length. The encoded data
        is a sequence of runs, where each run consists of a length byte
        followed by 1 to 128 bytes of data. If the length byte is in the
        range 0 to 127, the following length + 1 (1 to 128) bytes are
        copied literally during decompression. If length is in the range
        129 to 255, the following single byte is to be copied 257 - length
        (2 to 128) times during decompression. A length value of 128
        denotes EOD.
    """
    decoded = b''
    i = 0
    while i < len(data):
        length = data[i]
        if length == 128:
            break
        if length >= 0 and length < 128:
            for j in range(i+1,(i+1)+(length+1)):
                decoded+=int2byte(data[j])
            
            i = (i+1) + (length+1)
        if length > 128:
            run = int2byte(data[i+1])*(257-length)
            decoded+=run
            i = (i+1) + 1
    return decoded




import zlib

from ...utils import lzwdecode
from ...utils import ascii85decode, asciihexdecode, apply_png_predictor

from .constants import * 
from .types import * 
from .exceptions import * 
from .util import *
from .ccitt import ccittfaxdecode

class PDFStream(PDFObject):

    @classmethod
    def validated_stream(cls, x):
        x = resolve1(x)
        if not isinstance(x, cls):
            if settings.STRICT:
                raise PDFTypeError('PDFStream required: %r' % x)
            return PDFStream({}, b'')
        return x

    def __init__(self, attrs, rawdata, decipher=None):
        assert isinstance(attrs, dict), str(type(attrs))
        self.attrs = attrs
        self.rawdata = rawdata
        self.decipher = decipher
        self.data = None
        self.objid = None
        self.genno = None
        return

    def set_objid(self, objid, genno):
        self.objid = objid
        self.genno = genno
        return

    def __repr__(self):
        if self.data is None:
            assert self.rawdata is not None
            return '<PDFStream(%r): raw=%d, %r>' % (self.objid, len(self.rawdata), self.attrs)
        else:
            assert self.data is not None
            return '<PDFStream(%r): len=%d, %r>' % (self.objid, len(self.data), self.attrs)

    def __contains__(self, name):
        return name in self.attrs

    def __getitem__(self, name):
        return self.attrs[name]

    def get(self, name, default=None):
        return self.attrs.get(name, default)

    def get_any(self, names, default=None):
        for name in names:
            if name in self.attrs:
                return self.attrs[name]
        return default

    def get_filters(self):
        filters = self.get_any(('F', 'Filter'))
        params = self.get_any(('DP', 'DecodeParms', 'FDecodeParms'), {})
        if not filters:
            return []
        if not isinstance(filters, list):
            filters = [filters]
        if not isinstance(params, list):
            # Make sure the parameters list is the same as filters.
            params = [params] * len(filters)
        if settings.STRICT and len(params) != len(filters):
            raise PDFException("Parameters len filter mismatch")
        # resolve filter if possible
        _filters = []
        for fltr in filters:
            if hasattr(fltr, 'resolve'):
                fltr = fltr.resolve()[0]
            _filters.append(fltr)
        return list(zip(_filters, params)) #solves https://github.com/pdfminer/pdfminer.six/issues/15

    def decode(self):
        assert self.data is None and self.rawdata is not None, str((self.data, self.rawdata))
        data = self.rawdata
        if self.decipher:
            # Handle encryption
            data = self.decipher(self.objid, self.genno, data, self.attrs)
        filters = self.get_filters()
        if not filters:
            self.data = data
            self.rawdata = None
            return
        for (f,params) in filters:
            if f in LITERALS_FLATE_DECODE:
                # will get errors if the document is encrypted.
                try:
                    data = zlib.decompress(data)
                except zlib.error as e:
                    if settings.STRICT:
                        raise PDFException('Invalid zlib bytes: %r, %r' % (e, data))
                    data = b''
            elif f in LITERALS_LZW_DECODE:
                data = lzwdecode(data)
            elif f in LITERALS_ASCII85_DECODE:
                data = ascii85decode(data)
            elif f in LITERALS_ASCIIHEX_DECODE:
                data = asciihexdecode(data)
            elif f in LITERALS_RUNLENGTH_DECODE:
                data = rldecode(data)
            elif f in LITERALS_CCITTFAX_DECODE:
                data = ccittfaxdecode(data, params)
            elif f in LITERALS_DCT_DECODE:
                # This is probably a JPG stream - it does not need to be decoded twice.
                # Just return the stream to the user.
                pass
            elif f == LITERAL_CRYPT:
                # not yet..
                raise PDFNotImplementedError('/Crypt filter is unsupported')
            else:
                raise PDFNotImplementedError('Unsupported filter: %r' % f)
            # apply predictors
            if params and 'Predictor' in params:
                pred = int_value(params['Predictor'])
                if pred == 1:
                    # no predictor
                    pass
                elif 10 <= pred:
                    # PNG predictor
                    colors = int_value(params.get('Colors', 1))
                    columns = int_value(params.get('Columns', 1))
                    bitspercomponent = int_value(params.get('BitsPerComponent', 8))
                    data = apply_png_predictor(pred, colors, columns, bitspercomponent, data)
                else:
                    raise PDFNotImplementedError('Unsupported predictor: %r' % pred)
        self.data = data
        self.rawdata = None
        return

    def get_data(self):
        if self.data is None:
            self.decode()
        return self.data

    def get_rawdata(self):
        return self.rawdata


from pdfmajor.parser.PDFPage import PDFPage
from pdfmajor.parser.PDFStream import list_value

from .commands import PDFStateStack
from .commands import process_command_stream, prep_state

class PageInterpreter:
    class PDFInterpreterError(Exception): pass

    def __init__(self, page: PDFPage, page_num: int, font_cache: dict = {}):
        (x0, y0, x1, y1) = page.mediabox
        if page.rotate == 90:
            ctm = [0, -1, 1, 0, -y0, x1]
        elif page.rotate == 180:
            ctm = [-1, 0, 0, -1, x1, y1]
        elif page.rotate == 270:
            ctm = [0, 1, -1, 0, y1, -x0]
        else:
            ctm = [1, 0, 0, 1, -x0, -y0]

        # Locals
        self.font_cache = font_cache
        self.page = page
        self.page_num = page_num
        self.height = y1-y0
        self.width = x1-x0

        # Init State
        self.state: PDFStateStack = prep_state(
            PDFStateStack(), 
            ctm=ctm, 
            resources=page.resources, 
            font_cache=self.font_cache
        )

    
    def __iter__(self):
        for item in process_command_stream(
            streams=list_value(self.page.contents),
            font_cache=self.font_cache,
            state=self.state
        ):
            yield item

    def __repr__(self) -> str:
        return f"<Page:{self.page_num} width={self.width} height={self.height}/>"


from typing import List
from pdfmajor.utils import apply_matrix_pt, INF, Bbox, Point, MATRIX_IDENTITY, mult_matrix

from ..Curves import CurvePath
from ..PDFGraphicState import PDFGraphicState, PDFColor
from ..PDFTextState import PDFTextState

from ._base import LTItem, LTComponent, LTContainer
from .LTTextBlock import LTTextBlock
from .LTCharBlock import LTCharBlock, LTChar
from .LTCurves import LTCurve, LTLine, LTHorizontalLine, LTRect, LTVerticalLine
from .LTImage import LTImage
from .LTXObject import LTXObject
from .utils.textdecoder import decode_text_seq


def make_char_block(seq: bytearray, ctm: tuple, textstate: PDFTextState, color: PDFColor):
    (char_meta_datas, textstate, color) = decode_text_seq(seq, ctm, textstate, color)
    return LTCharBlock(
        chars=char_meta_datas,
        color=color,
        textstate=textstate
    )

def make_curve(ctm: tuple, gstate: PDFGraphicState, evenodd: bool, paths: List[CurvePath]):
    shape = ''.join(x.method.value for x in paths)
    if shape == 'ml':
        # horizontal/vertical line
        pt0 = paths[0].points[0]
        pt1 = paths[1].points[0]
        (x0, y0) = apply_matrix_pt(ctm, (pt0.x, pt0.y))
        (x1, y1) = apply_matrix_pt(ctm, (pt1.x, pt1.y))
        if x0 == x1:
            return LTVerticalLine(
                linewidth=gstate.linewidth, 
                paths=paths,
                bbox=Bbox( x0, y0, x1, y1 ),
                stroke=gstate.scolor, 
                fill=gstate.ncolor, 
                evenodd=evenodd, 
            )
        elif y0 == y1:
            return LTHorizontalLine(
                linewidth=gstate.linewidth, 
                paths=paths,
                bbox=Bbox( x0, y0, x1, y1 ),
                stroke=gstate.scolor, 
                fill=gstate.ncolor, 
                evenodd=evenodd, 
            )
        else:
            return LTLine(
                linewidth=gstate.linewidth, 
                paths=paths,
                bbox=Bbox( x0, y0, x1, y1 ),
                stroke=gstate.scolor, 
                fill=gstate.ncolor, 
                evenodd=evenodd, 
            )
    elif shape == 'mlllh':
        # rectangle
        pt0 = paths[0].points[0]
        pt1 = paths[1].points[0]
        pt2 = paths[2].points[0]
        pt3 = paths[3].points[0]
        (x0, y0) = apply_matrix_pt(ctm, (pt0.x, pt0.y))
        (x1, y1) = apply_matrix_pt(ctm, (pt1.x, pt1.y))
        (x2, y2) = apply_matrix_pt(ctm, (pt2.x, pt2.y))
        (x3, y3) = apply_matrix_pt(ctm, (pt3.x, pt3.y))
        if ((x0 == x1 and y1 == y2 and x2 == x3 and y3 == y0) or
            (y0 == y1 and x1 == x2 and y2 == y3 and x3 == x0)):
            return LTRect( 
                linewidth=gstate.linewidth, 
                paths=paths,
                bbox=Bbox(
                    x0=min(x0,x1,x2,x3),
                    x1=max(x0,x1,x2,x3),
                    y0=min(y0,y1,y2,y3),
                    y1=max(y0,y1,y2,y3),
                ),
                evenodd=evenodd, 
                stroke=gstate.scolor, 
                fill=gstate.ncolor,
            )
    else:
        (x0, y0, x1, y1) = (INF, INF, -INF, -INF)
        for path in paths:
            if path.method in [CurvePath.METHOD.MOVE_TO, CurvePath.METHOD.LINE_TO]:
                x0 = min(x0, path.points[0].x)
                y0 = min(y0, path.points[0].y)
                x1 = max(x1, path.points[0].x)
                y1 = max(y1, path.points[0].y)
            elif path.method in [CurvePath.METHOD.CURVE_BOTH_TO]:
                x0 = min(x0, path.points[2].x)
                y0 = min(y0, path.points[2].y)
                x1 = max(x1, path.points[2].x)
                y1 = max(y1, path.points[2].y)
            elif path.method in [CurvePath.METHOD.CURVE_NEXT_TO, CurvePath.METHOD.CURVE_FIRST_TO]:
                x0 = min(x0, path.points[1].x)
                y0 = min(y0, path.points[1].y)
                x1 = max(x1, path.points[1].x)
                y1 = max(y1, path.points[1].y)
        return LTCurve(
            linewidth=gstate.linewidth, 
            paths=paths,
            bbox=Bbox(x0,y0,x1,y1), 
            evenodd=evenodd,
            stroke=gstate.scolor, 
            fill=gstate.ncolor
        )

def make_image(obj, ctm):
    matrix = mult_matrix(MATRIX_IDENTITY, ctm)
    return LTImage(
        name=str(id(obj)),
        stream=obj,
        bbox=Bbox.from_points([
            apply_matrix_pt(matrix, (p, q))
            for (p, q) in ((0, 0), (1, 0), (0, 1), (1, 1))
        ])
    )

def make_xobject(obj, bbox, ctm, matrix, resources):
    (x, y, w, h) = bbox
    return LTXObject(
        name=str(id(obj)),
        bbox=Bbox.from_points([
            apply_matrix_pt(matrix, (p, q))
            for (p, q) in ((x, y), (x+w, y), (x, y+h), (x+w, y+h))
        ]),
        xobj_stream=obj,
        resources=resources,
        t_matrix = ctm
    )


from typing import List

from pdfmajor.parser.PDFContentParser import PDFContentParser, PDFStream
from pdfmajor.parser.PSStackParser import PSKeyword, keyword_name
from pdfmajor.utils import get_logger

from .commands import PDFCommands
from .state import PDFStateStack, PDFGraphicState, PDFTextState
from .state.PDFGraphicState import PDFColor, PDFColorSpace, PREDEFINED_COLORSPACE
from .state import PDFFont, get_font
from .state import LTTextBlock, LTCharBlock, LTChar
from .state import LTCurve, LTLine, LTHorizontalLine, LTVerticalLine, LTRect
from .state import LTImage
from .state import LTXObject
from .state import LTItem, LTComponent, LTContainer
from .utils import init_resources

log = get_logger('process_command_stream')

class CommandProcessorError(Exception): pass

def process_command_stream(streams: List[PDFStream], font_cache: dict = None, state: PDFStateStack = None):
    if state is None:
        state = PDFStateStack()
    if font_cache is None:
        font_cache = {}
    parser = PDFContentParser(streams)
    for obj in parser:
        if isinstance(obj, PSKeyword):
            name = keyword_name(obj)
            method = name.replace('*', '_a').replace('"', '_w').replace("'", '_q')
            if method in PDFCommands.commands.keys():
                func = PDFCommands.commands.get(method)
                nargs = func.__code__.co_argcount-1
                if nargs:
                    args = state.pop(nargs)
                    if len(args) == nargs:
                        func(state, *args)
                    else:
                        raise CommandProcessorError(f"Invalid Number of Args provided {method}")
                else:
                    func(state)
            else:
                log.warning('Unknown operator: %r (i.e. %r)' % (name, method))
        else:
            state.argstack.append(obj)
        
        for complete_item in state.complete_layout_items:
            if isinstance(complete_item, LTXObject):
                xobj_state = prep_state(
                    state=PDFStateStack(),
                    ctm=complete_item.t_matrix,
                    resources=complete_item.resources,
                    font_cache=font_cache
                )
                for item in process_command_stream(
                    [complete_item.stream], 
                    font_cache=font_cache, 
                    state=xobj_state
                ):
                    complete_item.add(item)
            yield complete_item
        state.complete_layout_items = []

def prep_state(state: PDFStateStack, ctm: tuple, resources: dict, font_cache: dict) -> PDFStateStack:
    state.t_matrix = ctm
    state.resources = resources

    # set some global states.
    state.graphics.ncolspace = state.graphics.scolspace = None 
    if state.colorspace_map:
        col_space = next(iter(state.colorspace_map.values()))
        state.graphics.ncolspace = state.graphics.scolspace = col_space
    init_resources(state, font_cache)
    return state


from typing import List

from ..interpreter import PDFInterpreter, PageInterpreter, logging
from ..interpreter import LTTextBlock, LTXObject


def convert_to_text(
    input_file_path: str, 
    output_file_path: str, 
    image_folder_path: str = None,
    dont_export_images: bool = False,
    codec: str = 'utf-8',
    maxpages: int = 0, 
    password: str = None, 
    caching: bool = True, 
    check_extractable: bool = True,
    pagenos: List[int] = None,
    debug_level: int = logging.WARNING,
):
    intepreter = PDFInterpreter(input_file_path, 
        maxpages=maxpages, 
        password=password, 
        caching=caching,
        check_extractable=check_extractable,
        pagenos=pagenos,
        debug_level=debug_level
    )
    with open(output_file_path, 'wb') as outfp:
        def process_container(container: LTXObject):
            for item in container:
                if isinstance(item, LTTextBlock):
                    for text in item:
                        for char in text:
                            outfp.write(char.get_text().encode(codec))
                elif isinstance(item, LTXObject):
                    process_container(item)
        for page in intepreter:
            outfp.write(f"========== [ page {page.page_num} ] ==========\n".encode(codec))
            process_container(page)
            outfp.write('\n'.encode(codec))



from .PDFColor import PDFColor, PDFColorSpace
from .PDFColorSpace import PREDEFINED_COLORSPACE

class PDFGraphicState(object):
    class InvalidOperation(Exception): pass

    def __init__(self):
        self.linewidth = 0
        self.linecap = None
        self.linejoin = None
        self.miterlimit = None
        self.dash = None
        self.intent = None
        self.flatness = None
        self.scolor:PDFColor = PDFColor(None)
        self.ncolor:PDFColor = PDFColor(None)
        self.scolspace:PDFColorSpace = None
        self.ncolspace:PDFColorSpace = None

    def set_stroke_color(self, colspace: PDFColorSpace, *values):
        if colspace is None:
            colspace = self.scolspace
        self.scolor = PDFColor(colspace, *values)

    def set_nostroke_color(self, colspace: PDFColorSpace, *values):
        if colspace is None:
            colspace = self.ncolspace
        self.ncolor = PDFColor(colspace, *values)

    def copy(self):
        obj = PDFGraphicState()
        obj.linewidth   = self.linewidth
        obj.linecap     = self.linecap
        obj.linejoin    = self.linejoin
        obj.miterlimit  = self.miterlimit
        obj.dash        = self.dash
        obj.intent      = self.intent
        obj.flatness    = self.flatness
        obj.scolor      = self.scolor.copy()
        obj.ncolor      = self.ncolor.copy()
        obj.scolspace   = self.scolspace
        obj.ncolspace   = self.ncolspace
        return obj

    def __repr__(self):
        return ('<PDFGraphicState linewidth=%r, linecap=%r, linejoin=%r, '
                ' miterlimit=%r, dash=%r, intent=%r, flatness=%r, '
                ' stroking color=%r, non stroking color=%r'
                '>' %
                (
                    self.linewidth, self.linecap, self.linejoin,
                    self.miterlimit, self.dash, self.intent, self.flatness,
                    str(self.scolor), str(self.ncolor)
                 )
        )



from typing import List
from .writers.html import HTMLMaker
from ..interpreter import PDFInterpreter, PageInterpreter, logging
from ..interpreter import LTImage, LTTextBlock, LTCharBlock, LTChar, LTCurve, LTXObject
from ..interpreter.commands import LTItem
from ..interpreter.commands.state import CurvePath, PDFColor

def convert_to_html(
    input_file_path: str, 
    output_file_path: str, 
    image_folder_path: str = None,
    dont_export_images: bool = False,
    codec: str = 'utf-8',
    maxpages: int = 0, 
    password: str = None, 
    caching: bool = True, 
    check_extractable: bool = True,
    pagenos: List[int] = None,
    debug_level: int = logging.WARNING,
):
    intepreter = PDFInterpreter(input_file_path, 
        maxpages=maxpages, 
        password=password, 
        caching=caching,
        check_extractable=check_extractable,
        pagenos=pagenos,
        debug_level=debug_level
    )
    with HTMLMaker(output_file_path, codec=codec) as html:
        with html.elm("html", nolineend=True):
            with html.elm("head"):
                meta_attr = {
                    'http-equiv': "Content-Type",
                    'content': "text/html",
                }
                if html.codec:
                    meta_attr['charset'] = html.codec
                html.singleton("meta", meta_attr)
                with html.elm('style'):
                    html.write('''.page { position: relative; overflow: hidden;border: 1px black solid;}''')
            with html.elm("body"):
                for page in intepreter:
                    render_page(html, page)

def render_page(html: HTMLMaker, ltpage: PageInterpreter):
    def render(item: LTItem):
        if isinstance(item, LTCurve):
            render_curve(html, item)
        elif isinstance(item, LTXObject):
            with html.elm('div', { "class": 'xobject' }, {
                'position': 'absolute', 
                'left': f'{item.x0}px', 
                'bottom': f'{item.y0}px', 
                'width': f'{item.width}px', 
                'height': f'{item.height}px',
            }):
                for child in item:
                    render(child)
        elif isinstance(item, LTImage):
            name = item.name
            # if html.imagewriter is not None:
            #     name = html.imagewriter.export_image(item)
            html.singleton('img', attrs={
                'src': name,
                'background': 'red',
                'width': f"{item.width}px", 
                'height': f"{item.height}px",
            })
        elif isinstance(item, LTChar):
            with html.elm('span', { 'class': 'char' }, {
                'width': item.width,
                'min-width': item.width,
                'max-width': item.width,
                'height': item.height,
                'min-height': item.height,
                'max-height': item.height,
            }):
                html.write(item.get_text())
        elif isinstance(item, LTTextBlock):
            for char in item:
                render(char)
        elif isinstance(item, LTCharBlock):
            with html.elm('span', {'class': 'char-block'}):
                with html.elm('p', { 'class': 'text-block' }, {
                    'position': 'absolute',
                    'left': f'{item.x0}',
                    'bottom': f'{item.y0}',
                    'font-size': f'{item.size}px',
                    'font-family': item.fontname,
                    'font-weight': item.font.font_weight,
                    'color': get_color(item.color, default="black"),
                    'width': item.width,
                    'min-width': item.width,
                    'max-width': item.width,
                    'height': item.height,
                    'min-height': item.height,
                    'max-height': item.height,
                    'transform': f'skew({item.font.italic_angle}deg, 0deg)',
                    'text-align': 'center',
                    'display': 'flex',
                    'flex-flow': 'row nowrap',
                    'justify-content': 'space-between',
                    # 'padding-left': item.font.leading
                }):
                    for char in item:
                        render(char)
    
    with html.elm('div', { "class": 'page', "id": f"page-{ltpage.page_num}" }, { 
        'width': f'{ltpage.width}px', 
        'height': f'{ltpage.height}px' 
    }):
        for child in ltpage:
            render(child)

def get_color(col: PDFColor, default: str = ""):
    if col.color_space is not None:
        if col.color_space.ncomponents == 1:
            if type(col.values[0]) in [float, int]:
                return f"rgba(0,0,0,{1 - float(col.values[0])})"
        elif col.color_space.ncomponents == 3:
            return f"rgb({','.join(map(str, col.values))})"
        elif col.color_space.ncomponents == 4:
            return f"cmyk({','.join(map(str, col.values))})"
        return f"{col.color_space.name}({','.join(map(str, col.values))})"
    else:
        return default

def render_curve(html: HTMLMaker, item: LTCurve):
        css = {
            'position': 'absolute', 
            'left': f'{item.x0}px', 
            'bottom': f'{item.y0}px', 
            'width': f'{item.width}px',
            'height': f'{item.height}px',
        }
        path = ""
        svg_attr = {
            'class': 'curve',
            'z-index': 2
        }
        path_css = {

        }

        # Compute ViewBox
        p_vals = { "x": [], "y": []  }

        for p in item.paths:
            for pt in p.points:
                p_vals['x'].append(pt.x)
                p_vals['y'].append(pt.y)

        mx_x = max([*p_vals['x'], 1])
        mn_x = min([*p_vals['x'], 0])
        mx_y = max([*p_vals['y'], 1])
        mn_y = min([*p_vals['y'], 0])

        if mx_x == mn_x:
            mn_x -= 1
        if mx_y == mn_y:
            mn_y -= 1

        dx = 1/( mx_x - mn_x )
        dy = (item.height/item.width)/( mx_y - mn_y )
        ROUND_VAL = 5
        svg_attr.update({
            # "viewBox": f"{min(p_vals['x'])} {min(p_vals['y'])} {max(p_vals['x'])} {max(p_vals['y'])}",
            "viewBox": f"{round(mn_x*dx, ROUND_VAL)} {round(mn_y*dy, ROUND_VAL)} {round(mx_x*dx , ROUND_VAL)} {round(mx_y*dy, ROUND_VAL)}",
        })
        
        # Compute shape
        for p in item.paths:
            if p.method == CurvePath.METHOD.MOVE_TO:
                path += f'M{" ".join( f"{round(dx*pt.x, ROUND_VAL)} {round(dy*pt.y, ROUND_VAL)}" for pt in p.points )} '
            if p.method == CurvePath.METHOD.LINE_TO:
                path += f'L{" ".join( f"{round(dx*pt.x, ROUND_VAL)} {round(dy*pt.y, ROUND_VAL)}" for pt in p.points )} '
            if p.method == CurvePath.METHOD.CLOSE_PATH:
                path += 'Z'
        
        stroke_col = get_color(item.stroke)
        fill_col = get_color(item.fill)
        if fill_col:
            path_css['fill'] = fill_col
        if stroke_col:
            path_css['stroke'] = stroke_col

        with html.elm('svg', svg_attr, css):
            html.singleton('path', {
                "d": path,
                **path_css
            }, {
                **path_css
            })



from pdfmajor.utils import mult_matrix, translate_matrix, Bbox, apply_matrix_pt, isnumber
from ....state import PDFTextState, PDFColor

def decode_text_seq(seq: bytearray, ctm: tuple, textstate: PDFTextState, color: PDFColor):
    textstate.matrix = mult_matrix(textstate.matrix, ctm)
    textstate.scaling = textstate.scaling * .01
    textstate.charspace = textstate.charspace * textstate.scaling
    textstate.wordspace = textstate.wordspace * textstate.scaling
    
    if textstate.font.is_multibyte():
        textstate.wordspace = 0
    dxscale = .001 * textstate.fontsize * textstate.scaling

    if textstate.font.is_vertical():
        return __render_string_along(1,
            seq, textstate, dxscale, color
        )
    else:
        return __render_string_along(0,
            seq, textstate, dxscale, color
        )

def __render_string_along(idx: int, seq: bytearray, 
    textstate: PDFTextState, dxscale: float, color: PDFColor):
    needcharspace = False
    char_meta_datas = []
    for obj in seq:
        if isnumber(obj):
            textstate.linematrix[idx] -= obj*dxscale
            needcharspace = True
        else:
            for cid in textstate.font.decode(obj):
                if needcharspace:
                    textstate.linematrix[idx] += textstate.charspace
                
                text = textstate.font.to_unichr(cid)
                assert isinstance(text, str), str(type(text))

                matrix = translate_matrix(textstate.matrix, textstate.linematrix)
                adv, bbox = __compute_char_bbox(matrix, cid, textstate)
                textstate.linematrix[idx] += adv

                char_meta_datas.append(
                    ( text, bbox )
                )
                if cid == 32 and textstate.wordspace:
                    textstate.linematrix[idx] += textstate.wordspace
                needcharspace = True
                
    return (char_meta_datas, textstate, color)

def __compute_char_bbox(matrix, char_id: int, textstate: PDFTextState):
    font = textstate.font
    fontsize = textstate.fontsize
    rise = textstate.rise
    scaling = textstate.scaling

    textwidth = font.char_width(char_id)
    textdisp = font.char_disp(char_id)
    adv = textwidth * fontsize * scaling

    # compute the boundary rectangle.
    if font.is_vertical():
        # vertical
        width = font.get_width() * fontsize
        (vx, vy) = textdisp
        if vx is None:
            vx = width * 0.5
        else:
            vx = vx * fontsize * .001
        vy = (1000 - vy) * fontsize * .001
        tx = -vx
        ty = vy + rise
        bll = (tx, ty+adv)
        bur = (tx+width, ty)
    else:
        # horizontal
        height = font.get_height() * fontsize
        descent = font.get_descent() * fontsize
        ty = descent + rise
        bll = (0, ty)
        bur = (adv, ty+height)
    (x0, y0) = apply_matrix_pt(matrix, bll)
    (x1, y1) = apply_matrix_pt(matrix, bur)
    if x1 < x0:
        (x0, x1) = (x1, x0)
    if y1 < y0:
        (y0, y1) = (y1, y0)
    
    return (adv, ((x0, y0), (x1, y1)))


import logging

logging.basicConfig(
    level=logging.WARNING
)

LOGGERS = {}

def set_log_level(debug_level: int):
    logging.basicConfig(level=debug_level)
    for _, logger in LOGGERS.items():
        logger.setLevel(debug_level)

def get_logger(logger_name: str) -> logging.Logger:
    if logger_name not in LOGGERS.keys():
        LOGGERS[logger_name] = logging.getLogger(logger_name)
    return LOGGERS[logger_name]


##  CFFFont
##  (Format specified in Adobe Technical Note: #5176
##   "The Compact Font Format Specification")
##
import struct
from io import BytesIO
from pdfmajor.utils import nunpack

NIBBLES = ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.', 'e', 'e-', None, '-')

def getdict(data):
    d = {}
    fp = BytesIO(data)
    stack = []
    while 1:
        c = fp.read(1)
        if not c:
            break
        b0 = ord(c)
        if b0 <= 21:
            d[b0] = stack
            stack = []
            continue
        if b0 == 30:
            s = ''
            loop = True
            while loop:
                b = ord(fp.read(1))
                for n in (b >> 4, b & 15):
                    if n == 15:
                        loop = False
                    else:
                        s += NIBBLES[n]
            value = float(s)
        elif 32 <= b0 and b0 <= 246:
            value = b0-139
        else:
            b1 = ord(fp.read(1))
            if 247 <= b0 and b0 <= 250:
                value = ((b0-247) << 8)+b1+108
            elif 251 <= b0 and b0 <= 254:
                value = -((b0-251) << 8)-b1-108
            else:
                b2 = ord(fp.read(1))
                if 128 <= b1:
                    b1 -= 256
                if b0 == 28:
                    value = b1 << 8 | b2
                else:
                    value = b1 << 24 | b2 << 16 | struct.unpack('>H', fp.read(2))[0]
        stack.append(value)
    return d


class CFFFont(object):

    STANDARD_STRINGS = (
      '.notdef', 'space', 'exclam', 'quotedbl', 'numbersign',
      'dollar', 'percent', 'ampersand', 'quoteright', 'parenleft',
      'parenright', 'asterisk', 'plus', 'comma', 'hyphen', 'period',
      'slash', 'zero', 'one', 'two', 'three', 'four', 'five', 'six',
      'seven', 'eight', 'nine', 'colon', 'semicolon', 'less', 'equal',
      'greater', 'question', 'at', 'A', 'B', 'C', 'D', 'E', 'F', 'G',
      'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
      'U', 'V', 'W', 'X', 'Y', 'Z', 'bracketleft', 'backslash',
      'bracketright', 'asciicircum', 'underscore', 'quoteleft', 'a',
      'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
      'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
      'braceleft', 'bar', 'braceright', 'asciitilde', 'exclamdown',
      'cent', 'sterling', 'fraction', 'yen', 'florin', 'section',
      'currency', 'quotesingle', 'quotedblleft', 'guillemotleft',
      'guilsinglleft', 'guilsinglright', 'fi', 'fl', 'endash',
      'dagger', 'daggerdbl', 'periodcentered', 'paragraph', 'bullet',
      'quotesinglbase', 'quotedblbase', 'quotedblright',
      'guillemotright', 'ellipsis', 'perthousand', 'questiondown',
      'grave', 'acute', 'circumflex', 'tilde', 'macron', 'breve',
      'dotaccent', 'dieresis', 'ring', 'cedilla', 'hungarumlaut',
      'ogonek', 'caron', 'emdash', 'AE', 'ordfeminine', 'Lslash',
      'Oslash', 'OE', 'ordmasculine', 'ae', 'dotlessi', 'lslash',
      'oslash', 'oe', 'germandbls', 'onesuperior', 'logicalnot', 'mu',
      'trademark', 'Eth', 'onehalf', 'plusminus', 'Thorn',
      'onequarter', 'divide', 'brokenbar', 'degree', 'thorn',
      'threequarters', 'twosuperior', 'registered', 'minus', 'eth',
      'multiply', 'threesuperior', 'copyright', 'Aacute',
      'Acircumflex', 'Adieresis', 'Agrave', 'Aring', 'Atilde',
      'Ccedilla', 'Eacute', 'Ecircumflex', 'Edieresis', 'Egrave',
      'Iacute', 'Icircumflex', 'Idieresis', 'Igrave', 'Ntilde',
      'Oacute', 'Ocircumflex', 'Odieresis', 'Ograve', 'Otilde',
      'Scaron', 'Uacute', 'Ucircumflex', 'Udieresis', 'Ugrave',
      'Yacute', 'Ydieresis', 'Zcaron', 'aacute', 'acircumflex',
      'adieresis', 'agrave', 'aring', 'atilde', 'ccedilla', 'eacute',
      'ecircumflex', 'edieresis', 'egrave', 'iacute', 'icircumflex',
      'idieresis', 'igrave', 'ntilde', 'oacute', 'ocircumflex',
      'odieresis', 'ograve', 'otilde', 'scaron', 'uacute',
      'ucircumflex', 'udieresis', 'ugrave', 'yacute', 'ydieresis',
      'zcaron', 'exclamsmall', 'Hungarumlautsmall', 'dollaroldstyle',
      'dollarsuperior', 'ampersandsmall', 'Acutesmall',
      'parenleftsuperior', 'parenrightsuperior', 'twodotenleader',
      'onedotenleader', 'zerooldstyle', 'oneoldstyle', 'twooldstyle',
      'threeoldstyle', 'fouroldstyle', 'fiveoldstyle', 'sixoldstyle',
      'sevenoldstyle', 'eightoldstyle', 'nineoldstyle',
      'commasuperior', 'threequartersemdash', 'periodsuperior',
      'questionsmall', 'asuperior', 'bsuperior', 'centsuperior',
      'dsuperior', 'esuperior', 'isuperior', 'lsuperior', 'msuperior',
      'nsuperior', 'osuperior', 'rsuperior', 'ssuperior', 'tsuperior',
      'ff', 'ffi', 'ffl', 'parenleftinferior', 'parenrightinferior',
      'Circumflexsmall', 'hyphensuperior', 'Gravesmall', 'Asmall',
      'Bsmall', 'Csmall', 'Dsmall', 'Esmall', 'Fsmall', 'Gsmall',
      'Hsmall', 'Ismall', 'Jsmall', 'Ksmall', 'Lsmall', 'Msmall',
      'Nsmall', 'Osmall', 'Psmall', 'Qsmall', 'Rsmall', 'Ssmall',
      'Tsmall', 'Usmall', 'Vsmall', 'Wsmall', 'Xsmall', 'Ysmall',
      'Zsmall', 'colonmonetary', 'onefitted', 'rupiah', 'Tildesmall',
      'exclamdownsmall', 'centoldstyle', 'Lslashsmall', 'Scaronsmall',
      'Zcaronsmall', 'Dieresissmall', 'Brevesmall', 'Caronsmall',
      'Dotaccentsmall', 'Macronsmall', 'figuredash', 'hypheninferior',
      'Ogoneksmall', 'Ringsmall', 'Cedillasmall', 'questiondownsmall',
      'oneeighth', 'threeeighths', 'fiveeighths', 'seveneighths',
      'onethird', 'twothirds', 'zerosuperior', 'foursuperior',
      'fivesuperior', 'sixsuperior', 'sevensuperior', 'eightsuperior',
      'ninesuperior', 'zeroinferior', 'oneinferior', 'twoinferior',
      'threeinferior', 'fourinferior', 'fiveinferior', 'sixinferior',
      'seveninferior', 'eightinferior', 'nineinferior',
      'centinferior', 'dollarinferior', 'periodinferior',
      'commainferior', 'Agravesmall', 'Aacutesmall',
      'Acircumflexsmall', 'Atildesmall', 'Adieresissmall',
      'Aringsmall', 'AEsmall', 'Ccedillasmall', 'Egravesmall',
      'Eacutesmall', 'Ecircumflexsmall', 'Edieresissmall',
      'Igravesmall', 'Iacutesmall', 'Icircumflexsmall',
      'Idieresissmall', 'Ethsmall', 'Ntildesmall', 'Ogravesmall',
      'Oacutesmall', 'Ocircumflexsmall', 'Otildesmall',
      'Odieresissmall', 'OEsmall', 'Oslashsmall', 'Ugravesmall',
      'Uacutesmall', 'Ucircumflexsmall', 'Udieresissmall',
      'Yacutesmall', 'Thornsmall', 'Ydieresissmall', '001.000',
      '001.001', '001.002', '001.003', 'Black', 'Bold', 'Book',
      'Light', 'Medium', 'Regular', 'Roman', 'Semibold',
    )

    class INDEX(object):

        def __init__(self, fp):
            self.fp = fp
            self.offsets = []
            (count, offsize) = struct.unpack('>HB', self.fp.read(3))
            for _ in range(count+1):
                self.offsets.append(nunpack(self.fp.read(offsize)))
            self.base = self.fp.tell()-1
            self.fp.seek(self.base+self.offsets[-1])
            return

        def __repr__(self):
            return '<INDEX: size=%d>' % len(self)

        def __len__(self): # pylint: disable=E0303
            return len(self.offsets)-1

        def __getitem__(self, i):
            self.fp.seek(self.base+self.offsets[i])
            return self.fp.read(self.offsets[i+1]-self.offsets[i])

        def __iter__(self):
            return iter(self[i] for i in range(len(self)))

    def __init__(self, name, fp):
        self.name = name
        self.fp = fp
        # Header
        (_major, _minor, hdrsize, _) = struct.unpack('BBBB', self.fp.read(4))
        self.fp.read(hdrsize-4)
        # Name INDEX
        self.name_index = self.INDEX(self.fp)
        # Top DICT INDEX
        self.dict_index = self.INDEX(self.fp)
        # String INDEX
        self.string_index = self.INDEX(self.fp)
        # Global Subr INDEX
        self.subr_index = self.INDEX(self.fp)
        # Top DICT DATA
        self.top_dict = getdict(self.dict_index[0])
        (charset_pos,) = self.top_dict.get(15, [0])
        (encoding_pos,) = self.top_dict.get(16, [0])
        (charstring_pos,) = self.top_dict.get(17, [0])
        # CharStrings
        self.fp.seek(charstring_pos)
        self.charstring = self.INDEX(self.fp)
        self.nglyphs = len(self.charstring)
        # Encodings
        self.code2gid = {}
        self.gid2code = {}
        self.fp.seek(encoding_pos)
        format = self.fp.read(1)
        if format == b'\x00':
            # Format 0
            (n,) = struct.unpack('B', self.fp.read(1))
            for (code, gid) in enumerate(struct.unpack('B'*n, self.fp.read(n))):
                self.code2gid[code] = gid
                self.gid2code[gid] = code
        elif format == b'\x01':
            # Format 1
            (n,) = struct.unpack('B', self.fp.read(1))
            code = 0
            for _ in range(n):
                (first, nleft) = struct.unpack('BB', self.fp.read(2))
                for gid in range(first, first+nleft+1):
                    self.code2gid[code] = gid
                    self.gid2code[gid] = code
                    code += 1
        else:
            raise ValueError('unsupported encoding format: %r' % format)
        # Charsets
        self.name2gid = {}
        self.gid2name = {}
        self.fp.seek(charset_pos)
        format = self.fp.read(1)
        if format == b'\x00':
            # Format 0
            n = self.nglyphs-1
            for (gid, sid) in enumerate(struct.unpack('>'+'H'*n, self.fp.read(2*n))):
                gid += 1
                name = self.getstr(sid)
                self.name2gid[name] = gid
                self.gid2name[gid] = name
        elif format == b'\x01':
            # Format 1
            (n,) = struct.unpack('B', self.fp.read(1))
            sid = 0
            for _ in range(n):
                (first, nleft) = struct.unpack('BB', self.fp.read(2))
                for gid in range(first, first+nleft+1):
                    name = self.getstr(sid)
                    self.name2gid[name] = gid
                    self.gid2name[gid] = name
                    sid += 1
        elif format == b'\x02':
            # Format 2
            assert False, str(('Unhandled', format))
        else:
            raise ValueError('unsupported charset format: %r' % format)
        #print self.code2gid
        #print self.name2gid
        #assert 0
        return

    def getstr(self, sid):
        if sid < len(self.STANDARD_STRINGS):
            return self.STANDARD_STRINGS[sid]
        return self.string_index[sid-len(self.STANDARD_STRINGS)]



from typing import List

from pdfmajor.utils import Bbox, Point, matrix2str
from pdfmajor.parser.PDFStream import PDFStream

from ._base import LTContainer

##  LTXObject
##
class LTXObject(LTContainer):

    def __init__(self, name: str, bbox: Bbox, xobj_stream: PDFStream, resources: dict, t_matrix: List[Point] ):
        LTContainer.__init__(self, bbox)
        self.name = name
        self.stream = xobj_stream
        self.t_matrix = t_matrix
        self.resources = resources

    def __repr__(self):
        return ('<%s(%s) %s matrix=%s>' %
                (self.__class__.__name__, self.name,
                 str(self.bbox), matrix2str(self.t_matrix)))



from typing import List

from tqdm import tqdm

from ..parser.PDFPage import PDFPage
from ..utils import set_log_level, get_logger, logging
from .commands.state import PDFStateStack, PDFColorSpace, PREDEFINED_COLORSPACE
from .commands import PDFCommands
from .commands import LTTextBlock, LTCharBlock, LTChar
from .commands import LTCurve, LTLine, LTHorizontalLine, LTVerticalLine, LTRect
from .commands import LTImage
from .commands import LTXObject
from .PageInterpreter import PageInterpreter

log = get_logger(__name__)

class PDFInterpreter:
    class AccessError(Exception): pass

    def __init__(self,
        input_file_path: str, 
        maxpages: int = 0, 
        caching: bool = True, 
        check_extractable: bool = True,
        password: str = None, 
        pagenos: List[int] = None,
        preload: bool = False,
        debug_level: int = logging.WARNING, 
    ):
        self.input_file_path = input_file_path
        self.maxpages = maxpages
        self.caching = caching
        self.check_extractable = check_extractable
        self.password = password
        self.pagenos = pagenos
        self.debug_level = debug_level
        self.__pages = []
        if preload:
            for _ in self.__load_pages():
                pass
        
    def __load_pages(self):
        set_log_level(self.debug_level)
        log.info("Opening file...")
        with open(self.input_file_path, 'rb') as input_file:
            font_cache = {}
            log.info("Parsing file...")
            pages = PDFPage.get_pages(
                input_file, 
                pagenos=self.pagenos, 
                maxpages=self.maxpages, 
                password=self.password, 
                caching=self.caching, 
                check_extractable=self.check_extractable
            )
            for page_num, page in tqdm(enumerate(pages), desc="Reading pages...", disable=self.debug_level > logging.INFO):
                self.__pages.append(PageInterpreter(page, page_num, font_cache))
                yield self.__pages[-1]
            log.info(f"Done Reading {len(self.__pages)} pages.")
        set_log_level(logging.WARNING)
        if len(self.__pages) == 0:
            raise self.AccessError("No pages found in pdf-file")
    
    def __iter__(self):
        if len(self.__pages) > 0:
            set_log_level(self.debug_level)
            for page in tqdm(self.__pages, desc="Reading pages from store...", disable=self.debug_level > logging.INFO):
                yield page
            set_log_level(logging.WARNING)
        else:
            for page in self.__load_pages():
                yield page


from .PDFFont import PDFFont, get_font
from pdfmajor.utils import MATRIX_IDENTITY

##  PDFTextState
##
class PDFTextState(object):
        
    def __init__(self):
        self.font: PDFFont = None
        self.fontsize: int = 0
        self.charspace: int = 0
        self.wordspace: int = 0
        self.scaling: int = 100
        self.leading: int = 0
        self.render: int = 0
        self.rise: int = 0
        self.matrix = MATRIX_IDENTITY
        self.linematrix = [0, 0]

    def __repr__(self):
        return ('<PDFTextState: font=%r, fontsize=%r, charspace=%r, wordspace=%r, '
                ' scaling=%r, leading=%r, render=%r, rise=%r, '
                ' matrix=%r, linematrix=%r>' %
                (self.font, self.fontsize, self.charspace, self.wordspace,
                 self.scaling, self.leading, self.render, self.rise,
                 self.matrix, self.linematrix))

    def copy(self):
        obj = self.__class__()
        obj.font = self.font
        obj.fontsize = self.fontsize
        obj.charspace = self.charspace
        obj.wordspace = self.wordspace
        obj.scaling = self.scaling
        obj.leading = self.leading
        obj.render = self.render
        obj.rise = self.rise
        obj.matrix = self.matrix
        obj.linematrix = self.linematrix
        return obj

    def reset_matrices(self):
        self.matrix = MATRIX_IDENTITY
        self.linematrix = [0, 0]

    @property
    def is_upright(self):
        return (0 < self.matrix[:,0].prod()*self.scaling and self.matrix[0,1:].prod() <= 0)


import logging

log = logging.getLogger(__name__)

from io import TextIOWrapper
from typing import List, Tuple, Union
from ...utils import int2byte

from .exceptions import PSEOF
from .constants import EOL, NONSPC, KWD, END_LITERAL, LIT, HEX
from .constants import SPC, HEX_PAIR
from .constants import END_NUMBER, END_STRING, END_KEYWORD, END_HEX_STRING
from .constants import OCT_STRING, ESC_STRING, KEYWORD_DICT_BEGIN, KEYWORD_DICT_END
from .types import PSLiteral

class PSBaseParser(object):

    """Most basic PostScript parser that performs only tokenization.
    """
    BUFSIZ = 4096

    def __init__(self, fp: TextIOWrapper):
        self.fp = fp
        self.seek(0)
        return

    def __repr__(self):
        return '<%s: %r, bufpos=%d>' % (self.__class__.__name__, self.fp, self.bufpos)

    def flush(self):
        return

    def close(self):
        self.flush()
        return

    def tell(self):
        return self.bufpos+self.charpos

    def poll(self, pos=None, n=80):
        pos0 = self.fp.tell()
        if not pos:
            pos = self.bufpos+self.charpos
        self.fp.seek(pos)
        log.info('poll(%d): %r', pos, self.fp.read(n))
        self.fp.seek(pos0)
        return

    def seek(self, pos: int):
        """Seeks the parser to the given position.
        """
        log.debug('seek: %r', pos)
        self.fp.seek(pos)
        # reset the status for nextline()
        self.bufpos: int = pos
        self.buf: bytes = b''
        self.charpos: int = 0
        # reset the status for nexttoken()
        self._current_parse_func = self._parse_main
        self._curtoken: bytes = b''
        self._curtokenpos: int = 0
        self._tokens: List[Tuple[int, Union[int, float, bool, str, bytes, PSLiteral]]] = []
        return

    def fillbuf(self):
        if self.charpos < len(self.buf):
            return
        # fetch next chunk.
        self.bufpos = self.fp.tell()
        self.buf = self.fp.read(self.BUFSIZ)
        if not self.buf:
            raise PSEOF('Unexpected EOF')
        self.charpos = 0
        return

    def nextline(self):
        """Fetches a next line that ends either with \\r or \\n.
        """
        linebuf = b''
        linepos = self.bufpos + self.charpos
        eol = False
        while 1:
            self.fillbuf()
            if eol:
                c = self.buf[self.charpos:self.charpos+1]
                # handle b'\r\n'
                if c == b'\n':
                    linebuf += c
                    self.charpos += 1
                break
            m = EOL.search(self.buf, self.charpos)
            if m:
                linebuf += self.buf[self.charpos:m.end(0)]
                self.charpos = m.end(0)
                if linebuf[-1:] == b'\r':
                    eol = True
                else:
                    break
            else:
                linebuf += self.buf[self.charpos:]
                self.charpos = len(self.buf)
        log.debug('nextline: %r, %r', linepos, linebuf)
        return (linepos, linebuf)

    def revreadlines(self):
        """Fetches a next line backword.

        This is used to locate the trailers at the end of a file.
        """
        self.fp.seek(0, 2)
        pos = self.fp.tell()
        buf = b''
        while 0 < pos:
            prevpos = pos
            pos = max(0, pos-self.BUFSIZ)
            self.fp.seek(pos)
            s = self.fp.read(prevpos-pos)
            if not s:
                break
            while 1:
                n = max(s.rfind(b'\r'), s.rfind(b'\n'))
                if n == -1:
                    buf = s + buf
                    break
                yield s[n:] + buf
                s = s[:n]
                buf = b''
        return

    def _parse_main(self, s, i):
        m = NONSPC.search(s, i)
        if not m:
            return len(s)
        j = m.start(0)
        c = s[j:j+1]
        self._curtokenpos = self.bufpos+j
        if c == b'%':
            self._curtoken = b'%'
            self._current_parse_func = self._parse_comment
            return j+1
        elif c == b'/':
            self._curtoken = b''
            self._current_parse_func = self._parse_literal
            return j+1
        elif c in b'-+' or c.isdigit():
            self._curtoken = c
            self._current_parse_func = self._parse_number
            return j+1
        elif c == b'.':
            self._curtoken = c
            self._current_parse_func = self._parse_float
            return j+1
        elif c.isalpha():
            self._curtoken = c
            self._current_parse_func = self._parse_keyword
            return j+1
        elif c == b'(':
            self._curtoken = b''
            self.paren = 1
            self._current_parse_func = self._parse_string
            return j+1
        elif c == b'<':
            self._curtoken = b''
            self._current_parse_func = self._parse_wopen
            return j+1
        elif c == b'>':
            self._curtoken = b''
            self._current_parse_func = self._parse_wclose
            return j+1
        else:
            self._add_token(KWD(c))
            return j+1

    def _add_token(self, obj):
        self._tokens.append((self._curtokenpos, obj))
        return

    def _parse_comment(self, s, i):
        m = EOL.search(s, i)
        if not m:
            self._curtoken += s[i:]
            return len(s)
        j = m.start(0)
        self._curtoken += s[i:j]
        self._current_parse_func = self._parse_main
        # We ignore comments.
        #self._tokens.append(self._curtoken)
        return j

    def _parse_literal(self, s, i):
        m = END_LITERAL.search(s, i)
        if not m:
            self._curtoken += s[i:]
            return len(s)
        j = m.start(0)
        self._curtoken += s[i:j]
        c = s[j:j+1]
        if c == b'#':
            self.hex = b''
            self._current_parse_func = self._parse_literal_hex
            return j+1
        try:
            self._curtoken=str(self._curtoken,'utf-8')
        except:
            pass
        self._add_token(LIT(self._curtoken))
        self._current_parse_func = self._parse_main
        return j

    def _parse_literal_hex(self, s, i):
        c = s[i:i+1]
        if HEX.match(c) and len(self.hex) < 2:
            self.hex += c
            return i+1
        if self.hex:
            self._curtoken += int2byte(int(self.hex, 16))
        self._current_parse_func = self._parse_literal
        return i

    def _parse_number(self, s, i):
        m = END_NUMBER.search(s, i)
        if not m:
            self._curtoken += s[i:]
            return len(s)
        j = m.start(0)
        self._curtoken += s[i:j]
        c = s[j:j+1]
        if c == b'.':
            self._curtoken += c
            self._current_parse_func = self._parse_float
            return j+1
        try:
            self._add_token(int(self._curtoken))
        except ValueError:
            pass
        self._current_parse_func = self._parse_main
        return j

    def _parse_float(self, s, i):
        m = END_NUMBER.search(s, i)
        if not m:
            self._curtoken += s[i:]
            return len(s)
        j = m.start(0)
        self._curtoken += s[i:j]
        try:
            self._add_token(float(self._curtoken))
        except ValueError:
            pass
        self._current_parse_func = self._parse_main
        return j

    def _parse_keyword(self, s, i):
        m = END_KEYWORD.search(s, i)
        if not m:
            self._curtoken += s[i:]
            return len(s)
        j = m.start(0)
        self._curtoken += s[i:j]
        if self._curtoken == b'true':
            token = True
        elif self._curtoken == b'false':
            token = False
        else:
            token = KWD(self._curtoken)
        self._add_token(token)
        self._current_parse_func = self._parse_main
        return j

    def _parse_string(self, s, i):
        m = END_STRING.search(s, i)
        if not m:
            self._curtoken += s[i:]
            return len(s)
        j = m.start(0)
        self._curtoken += s[i:j]
        c = s[j:j+1]
        if c == b'\\':
            self.oct = b''
            self._current_parse_func = self._parse_string_1
            return j+1
        if c == b'(':
            self.paren += 1
            self._curtoken += c
            return j+1
        if c == b')':
            self.paren -= 1
            if self.paren:  # WTF, they said balanced parens need no special treatment.
                self._curtoken += c
                return j+1
        self._add_token(self._curtoken)
        self._current_parse_func = self._parse_main
        return j+1

    def _parse_string_1(self, s, i):
        c = s[i:i+1]
        if OCT_STRING.match(c) and len(self.oct) < 3:
            self.oct += c
            return i+1
        if self.oct:
            self._curtoken += int2byte(int(self.oct, 8))
            self._current_parse_func = self._parse_string
            return i
        if c in ESC_STRING:
            self._curtoken += int2byte(ESC_STRING[c])
        self._current_parse_func = self._parse_string
        return i+1

    def _parse_wopen(self, s, i):
        c = s[i:i+1]
        if c == b'<':
            self._add_token(KEYWORD_DICT_BEGIN)
            self._current_parse_func = self._parse_main
            i += 1
        else:
            self._current_parse_func = self._parse_hexstring
        return i

    def _parse_wclose(self, s, i):
        c = s[i:i+1]
        if c == b'>':
            self._add_token(KEYWORD_DICT_END)
            i += 1
        self._current_parse_func = self._parse_main
        return i

    def _parse_hexstring(self, s, i):
        m = END_HEX_STRING.search(s, i)
        if not m:
            self._curtoken += s[i:]
            return len(s)
        j = m.start(0)
        self._curtoken += s[i:j]
        token = HEX_PAIR.sub(lambda m: int2byte(int(m.group(0), 16)),SPC.sub(b'', self._curtoken))
        self._add_token(token)
        self._current_parse_func = self._parse_main
        return j

    def nexttoken(self):
        while not self._tokens:
            self.fillbuf()
            self.charpos = self._current_parse_func(self.buf, self.charpos)
        token = self._tokens.pop(0)
        log.debug('nexttoken: %r', token)
        return token



from json import dumps as json_dump
from io import TextIOWrapper
from contextlib import contextmanager     

@contextmanager
def JSONMaker(file_path: str, codec: str = 'utf-8'):
    with open(file_path, 'wb') as outf:
        with JSONMakerObject(outf, codec=codec) as obj:
            yield obj


class JSONMakerObject:
    class AccessError(Exception): pass
    def __init__(self, outfile: TextIOWrapper, codec: str = 'utf-8', levels_deep: int = 0):
        self.outfile: TextIOWrapper = outfile
        self.codec = codec
        self.levels_deep = levels_deep
        self.keys = set()

    def __enter__(self):
        self.write_raw('{')
        self.levels_deep += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.levels_deep -= 1
        self.write('}', prefix='\n')
    
    def write_raw(self, text):
        if self.outfile is not None:
            if self.codec:
                text = text.encode(self.codec)
            self.outfile.write(text)
        else:
            raise self.AccessError("Attempting to write to file without an open connection")
    
    def write(self, text: str, deep_space: str = '  ', prefix:str = ''):
        text = prefix + (deep_space*self.levels_deep) + text
        self.write_raw(text)

    def write_prop(self, key: str, val: str = ""):
        if key not in self.keys:
            self.keys.add(key)
            if len(self.keys) > 1:
                self.write(f'"{key}": {val}', prefix=',\n')
            else:
                self.write(f'"{key}": {val}', prefix='\n')
 
    def number(self, key: str, val: float or int):
        self.write_prop(key, val)
    
    def string(self, key: str, val: str):
        self.write_prop(key, f'"{val}"')
    
    @contextmanager
    def object(self, key: str):
        self.write_prop(key)
        with JSONMakerObject(self.outfile, codec=self.codec, levels_deep=self.levels_deep) as child:
            yield child
    
    @contextmanager
    def array(self, key: str):
        self.write_prop(key)
        with JSONMakerArray(self.outfile, codec=self.codec, levels_deep=self.levels_deep) as child:
            yield child
    
    def place_object(self, key: str, obj: dict):
        self.write_prop(key, json_dump(obj))
    
    

class JSONMakerArray:
    class AccessError(Exception): pass
    def __init__(self, outfile: TextIOWrapper, codec: str = 'utf-8', levels_deep: int = 0):
        self.outfile: TextIOWrapper = outfile
        self.codec = codec
        self.levels_deep = levels_deep
        self.childrencount = 0

    def __enter__(self):
        self.write_raw('[')
        self.levels_deep += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.levels_deep -= 1
        self.write(']', prefix='\n')
    
    def write_raw(self, text):
        if self.outfile is not None:
            if self.codec:
                text = text.encode(self.codec)
            self.outfile.write(text)
        else:
            raise self.AccessError("Attempting to write to file without an open connection")
    
    def write(self, text: str, deep_space: str = '  ', prefix:str = ''):
        text = prefix + (deep_space*self.levels_deep) + text
        self.write_raw(text)

    def write_prop(self, val: str = ""):
        self.childrencount += 1
        if self.childrencount > 1:
            self.write(f'{val}', prefix=',\n')
        else:
            self.write(f'{val}', prefix='\n')

    def number(self, val: float or int):
        self.write_prop(val)
    
    def string(self, val: str):
        self.write_prop(f'"{val}"')
    
    @contextmanager
    def object(self):
        self.write_prop()
        with JSONMakerObject(self.outfile, codec=self.codec, levels_deep=self.levels_deep) as child:
            yield child
    
    @contextmanager
    def array(self):
        self.write_prop()
        with JSONMakerArray(self.outfile, codec=self.codec, levels_deep=self.levels_deep) as child:
            yield child
    
    def place_object(self, obj: dict):
        self.write_prop(json_dump(obj))
    
    


class PDFItem: 
    def __repr__(self) -> str:
        return f"""<{self.__class__.__name__} {' '.join([
            f'{name}="{val}"'
            for name, val in vars(self).items()
        ])}/>"""



from pdfmajor.parser.PDFParser import PDFStream
from pdfmajor.parser.PDFStream import list_value, dict_value, resolve1
from pdfmajor.parser.PDFStream.PDFObjRef import PDFObjRef
from pdfmajor.parser.PSStackParser import literal_name
from pdfmajor.parser.constants import LITERAL_PDF, LITERAL_TEXT

from .state import PDFColorSpace, PREDEFINED_COLORSPACE
from .state import PDFStateStack, get_font

def get_colorspace(spec):
    if isinstance(spec, list):
        name = literal_name(spec[0])
    else:
        name = literal_name(spec)
    if name == 'ICCBased' and isinstance(spec, list) and 2 <= len(spec):
        return PDFColorSpace(name, PDFStream.validated_stream(spec[1])['N'])
    elif name == 'DeviceN' and isinstance(spec, list) and 2 <= len(spec):
        return PDFColorSpace(name, len(list_value(spec[1])))
    else:
        return PREDEFINED_COLORSPACE.get(name)

def get_procset(procs):
    for proc in procs:
        if proc is LITERAL_PDF:
            pass
        elif proc is LITERAL_TEXT:
            pass
        else:
            #raise PDFResourceError('ProcSet %r is not supported.' % proc)
            pass
    return

def init_resources(state: PDFStateStack, font_cache: dict = {}):
    for (k, v) in iter(dict_value(state.resources).items()):
        # log.debug('Resource: %r: %r', k, v)
        if k == 'Font':
            for (fontid, spec) in iter(dict_value(v).items()):
                objid = None
                if isinstance(spec, PDFObjRef):
                    objid = spec.objid
                spec = dict_value(spec)
                state.fontmap[fontid] = get_font(objid, spec, font_cache)
        elif k == 'ColorSpace':
            for (csid, spec) in iter(dict_value(v).items()):
                state.colorspace_map[csid] = get_colorspace(resolve1(spec))
        elif k == 'ProcSet':
            get_procset(list_value(v))
        elif k == 'XObject':
            for (xobjid, xobjstrm) in iter(dict_value(v).items()):
                state.xobjmap[xobjid] = xobjstrm


from typing import List

from .writers.json import JSONMaker, JSONMakerObject, JSONMakerArray
from ..interpreter import PDFInterpreter, PageInterpreter, logging
from ..interpreter import LTImage, LTTextBlock, LTCharBlock, LTCurve, LTXObject
from ..interpreter.commands import LTItem
from ..interpreter.commands import LTRect, LTLine
from ..interpreter.commands.state import CurvePath, PDFColor
from ..parser.PSStackParser import PSLiteral
from ..utils import enc

def normalize_color_values(values: list):
    ret = []
    for v in values:
        if isinstance(v, PSLiteral):
            ret.append( enc(str(v.name)).decode() )
        else:
            ret.append( enc(str(v)).decode() )
    return ret

def convert_to_json(
    input_file_path: str, 
    output_file_path: str, 
    image_folder_path: str = None,
    dont_export_images: bool = False,
    codec: str = 'utf-8',
    maxpages: int = 0, 
    password: str = None, 
    caching: bool = True, 
    check_extractable: bool = True,
    pagenos: List[int] = None,
    debug_level: int = logging.WARNING,
):
    intepreter = PDFInterpreter(input_file_path, 
        maxpages=maxpages, 
        password=password, 
        caching=caching,
        check_extractable=check_extractable,
        pagenos=pagenos,
        debug_level=debug_level
    )
    with JSONMaker(output_file_path, codec=codec) as json:
        with json.array("pages") as arr:
            for page in intepreter:
                with arr.object() as page_obj:
                    render_page(page_obj, page)

def render_page(json: JSONMakerObject, ltpage: PageInterpreter):  
    json.number("id", ltpage.page_num)
    json.number("width", ltpage.width)
    json.number("height", ltpage.height)
    with json.array("children") as chil_arr:
        for child in ltpage:
            with chil_arr.object() as chil_obj:
                render(chil_obj, child)

def render(json: JSONMakerObject, item: LTItem):
        if isinstance(item, LTCurve):
            place_curve(json, item)
        elif isinstance(item, LTXObject):
            json.string("type", 'xobject')
            json.number("x0", item.x0)
            json.number("x1", item.x1)
            json.number("y0", item.y0)
            json.number("y1", item.y1)
            with json.array("children") as chil_arr:
                for child in item:
                    with chil_arr.object() as chil_obj:
                        render(chil_obj, child)
        elif isinstance(item, LTImage):
            # if json.imagewriter is not None:
            #     name = json.imagewriter.export_image(item)
            json.string("type", 'img')
            json.string("src", enc(item.name, None))
            json.number("x0", item.x0)
            json.number("x1", item.x1)
            json.number("y0", item.y0)
            json.number("y1", item.y1)
        elif isinstance(item, LTTextBlock):
            place_text_block(json, item)

def place_text_block(json: JSONMakerObject, text_block: LTTextBlock):
    json.string("type", 'text-block')
    json.number("x0", text_block.x0)
    json.number("x1", text_block.x1)
    json.number("y0", text_block.y0)
    json.number("y1", text_block.y1)
    with json.array("char_blocks") as cb_arr:
        for cb in text_block:
            with cb_arr.object() as cb_obj:
                place_char_block(cb_obj, cb)

def place_char_block(json: JSONMakerObject, char_block: LTCharBlock):
    json.number("size", char_block.size)
    json.number("x0", char_block.x0)
    json.number("x1", char_block.x1)
    json.number("y0", char_block.y0)
    json.number("y1", char_block.y1)
    if char_block.color.color_space is not None:
        json.place_object("color", {
            "type": enc(char_block.color.color_space.name).decode(),
            "value": normalize_color_values(char_block.color.values)
        })
    for key, value in char_block.font.descriptor.items():
        if key != "Type" and "FontFile" not in key:
            json.string(key, value)
    json.string("text", 
        "".join([
            char.get_text()
            for char in char_block
        ])
    )

def place_curve(json: JSONMakerObject, item: LTCurve):
    json.number("x0", item.x0)
    json.number("x1", item.x1)
    json.number("y0", item.y0)
    json.number("y1", item.y1)
    if item.fill.color_space is not None:
        json.place_object("fill", {
            "type": enc(item.fill.color_space.name).decode(),
            "value": normalize_color_values(item.fill.values)
        })
    if item.stroke.color_space is not None:
        json.place_object("stroke", {
            "type": enc(item.stroke.color_space.name).decode(),
            "value": normalize_color_values(item.stroke.values)
        })
                    
    if isinstance(item, LTRect):
        json.string("type", 'rect')
    elif isinstance(item, LTLine):
        json.string("type", 'line')
    else:
        json.string("type", 'curve')
        with json.array('paths') as paths_arr:
            for path in item.paths:
                paths_arr.place_object({
                    "type": path.method.name,
                    "points": [
                        {"x": point.x, "y": point.y}
                        for point in path.points
                    ]
                })


from enum import Enum
from typing import List

class CurvePoint:
    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y
    
    def __repr__(self) -> str:
        return f'<CurvePoint ({self.x}, {self.y})/>'

class CurveMethod(Enum):
    MOVE_TO='m'
    LINE_TO='l'
    CURVE_BOTH_TO='c'
    CURVE_NEXT_TO='v'
    CURVE_FIRST_TO='y'
    CLOSE_PATH='h'

class CurvePath:
    METHOD = CurveMethod
    def __init__(self, method: CurveMethod, *points: List[CurvePoint]):
        self.method = method
        self.points = points
    
    def __repr__(self) -> str:
        return f'<CurvePath {self.method} path=[{",".join(map(repr, self.points))}]/>'


from io import BytesIO

from pdfmajor.parser.PSStackParser import literal_name
from pdfmajor.parser.PDFStream import int_value
from pdfmajor.parser.PDFStream import num_value
from pdfmajor.parser.PDFStream import list_value
from pdfmajor.parser.PDFStream import dict_value
from pdfmajor.parser.PDFStream import PDFStream
from pdfmajor.parser.PDFStream import resolve1
from pdfmajor.parser.cmapdb import CMap, CMapDB, CMapParser
from pdfmajor.parser.cmapdb import FileUnicodeMap
from pdfmajor.utils import settings, apply_matrix_norm

from .PDFFont import PDFFont, PDFSimpleFont
from .util import FontMetricsDB, get_widths, get_widths2
from .Type1FontHeaderParser import Type1FontHeaderParser
from .TrueTypeFont import TrueTypeFont

# PDFType1Font
class PDFType1Font(PDFSimpleFont):

    def __init__(self, spec):
        try:
            self.basefont = literal_name(spec['BaseFont'])
        except KeyError:
            if settings.STRICT:
                raise self.PDFFontError('BaseFont is missing')
            self.basefont = 'unknown'
        try:
            (descriptor, widths) = FontMetricsDB.get_metrics(self.basefont)
        except KeyError:
            descriptor = dict_value(spec.get('FontDescriptor', {}))
            firstchar = int_value(spec.get('FirstChar', 0))
            #lastchar = int_value(spec.get('LastChar', 255))
            widths = list_value(spec.get('Widths', [0]*256))
            widths = dict((i+firstchar, w) for (i, w) in enumerate(widths))
        PDFSimpleFont.__init__(self, descriptor, widths, spec)
        if 'Encoding' not in spec and 'FontFile' in descriptor:
            # try to recover the missing encoding info from the font file.
            self.fontfile = PDFStream.validated_stream(descriptor.get('FontFile'))
            length1 = int_value(self.fontfile['Length1'])
            data = self.fontfile.get_data()[:length1]
            parser = Type1FontHeaderParser(BytesIO(data))
            self.cid2unicode = parser.get_encoding()
        return

    def __repr__(self):
        return '<PDFType1Font: basefont=%r>' % self.basefont


# PDFTrueTypeFont
class PDFTrueTypeFont(PDFType1Font):

    def __repr__(self):
        return '<PDFTrueTypeFont: basefont=%r>' % self.basefont


# PDFType3Font
class PDFType3Font(PDFSimpleFont):

    def __init__(self, spec):
        firstchar = int_value(spec.get('FirstChar', 0))
        #lastchar = int_value(spec.get('LastChar', 0))
        widths = list_value(spec.get('Widths', [0]*256))
        widths = dict((i+firstchar, w) for (i, w) in enumerate(widths))
        if 'FontDescriptor' in spec:
            descriptor = dict_value(spec['FontDescriptor'])
        else:
            descriptor = {'Ascent': 0, 'Descent': 0,
                          'FontBBox': spec['FontBBox']}
        PDFSimpleFont.__init__(self, descriptor, widths, spec)
        self.matrix = tuple(list_value(spec.get('FontMatrix')))
        (_, self.descent, _, self.ascent) = self.bbox
        (self.hscale, self.vscale) = apply_matrix_norm(self.matrix, (1, 1))
        return

    def __repr__(self):
        return '<PDFType3Font>'


# PDFCIDFont
class PDFCIDFont(PDFFont):

    def __init__(self, spec, strict=settings.STRICT):
        try:
            self.basefont = literal_name(spec['BaseFont'])
        except KeyError:
            if strict:
                raise self.PDFFontError('BaseFont is missing')
            self.basefont = 'unknown'
        self.cidsysteminfo = dict_value(spec.get('CIDSystemInfo', {}))
        self.cidcoding = '%s-%s' % (resolve1(self.cidsysteminfo.get('Registry', b'unknown')).decode("latin1"),
                                    resolve1(self.cidsysteminfo.get('Ordering', b'unknown')).decode("latin1"))
        try:
            name = literal_name(spec['Encoding'])
        except KeyError:
            if strict:
                raise self.PDFFontError('Encoding is unspecified')
            name = 'unknown'
        try:
            self.cmap = CMapDB.get_cmap(name)
        except CMapDB.CMapNotFound as e:
            if strict:
                raise self.PDFFontError(e)
            self.cmap = CMap()
        try:
            descriptor = dict_value(spec['FontDescriptor'])
        except KeyError:
            if strict:
                raise self.PDFFontError('FontDescriptor is missing')
            descriptor = {}
        ttf = None
        if 'FontFile2' in descriptor:
            self.fontfile = PDFStream.validated_stream(descriptor.get('FontFile2'))
            ttf = TrueTypeFont(self.basefont,
                               BytesIO(self.fontfile.get_data()))
        self.unicode_map = None
        if 'ToUnicode' in spec:
            strm = PDFStream.validated_stream(spec['ToUnicode'])
            self.unicode_map = FileUnicodeMap()
            CMapParser(self.unicode_map, BytesIO(strm.get_data())).run()
        elif self.cidcoding in ('Adobe-Identity', 'Adobe-UCS'):
            if ttf:
                try:
                    self.unicode_map = ttf.create_unicode_map()
                except TrueTypeFont.CMapNotFound:
                    pass
        else:
            try:
                self.unicode_map = CMapDB.get_unicode_map(self.cidcoding, self.cmap.is_vertical())
            except CMapDB.CMapNotFound as e:
                pass

        self.vertical = self.cmap.is_vertical()
        if self.vertical:
            # writing mode: vertical
            widths = get_widths2(list_value(spec.get('W2', [])))
            self.disps = dict((cid, (vx, vy)) for (cid, (_, (vx, vy))) in iter(widths.items()))
            (vy, w) = spec.get('DW2', [880, -1000])
            self.default_disp = (None, vy)
            widths = dict((cid, w) for (cid, (w, _)) in iter(widths.items()))
            default_width = w
        else:
            # writing mode: horizontal
            self.disps = {}
            self.default_disp = 0
            widths = get_widths(list_value(spec.get('W', [])))
            default_width = spec.get('DW', 1000)
        PDFFont.__init__(self, descriptor, widths, default_width=default_width)
        return

    def __repr__(self):
        return '<PDFCIDFont: basefont=%r, cidcoding=%r>' % (self.basefont, self.cidcoding)

    def is_vertical(self):
        return self.vertical

    def is_multibyte(self):
        return True

    def decode(self, bytes):
        return self.cmap.decode(bytes)

    def char_disp(self, cid):
        "Returns an integer for horizontal fonts, a tuple for vertical fonts."
        return self.disps.get(cid, self.default_disp)

    def to_unichr(self, cid):
        try:
            if not self.unicode_map:
                raise KeyError(cid)
            return self.unicode_map.get_unichr(cid)
        except KeyError:
            raise self.PDFUnicodeNotDefined(self.cidcoding, cid)



from ._base import PDFItem

class PDFXObject(PDFItem):
    def __init__(self, xobj_strem, bbox, resources, ctm):
        self.stream = xobj_strem
        self.bbox = bbox
        self.resources = resources
        self.ctm = ctm
        self.children = []


##  TrueTypeFont
##
import struct
from pdfmajor.parser.cmapdb import FileUnicodeMap

class TrueTypeFont(object):

    class CMapNotFound(Exception):
        pass

    def __init__(self, name, fp):
        self.name = name
        self.fp = fp
        self.tables = {}
        self.fonttype = fp.read(4)
        try:
            (ntables, _1, _2, _3) = struct.unpack('>HHHH', fp.read(8))
            for _ in range(ntables):
                (name, _, offset, length) = struct.unpack('>4sLLL', fp.read(16))
                self.tables[name] = (offset, length)
        except struct.error:
            # Do not fail if there are not enough bytes to read. Even for
            # corrupted PDFs we would like to get as much information as
            # possible, so continue.
            pass
        return

    def create_unicode_map(self):
        if 'cmap' not in self.tables:
            raise TrueTypeFont.CMapNotFound
        (base_offset, _) = self.tables['cmap']
        fp = self.fp
        fp.seek(base_offset)
        (_, nsubtables) = struct.unpack('>HH', fp.read(4))
        subtables = []
        for i in range(nsubtables):
            subtables.append(struct.unpack('>HHL', fp.read(8)))
        char2gid = {}
        # Only supports subtable type 0, 2 and 4.
        for (_1, _2, st_offset) in subtables:
            fp.seek(base_offset+st_offset)
            (fmttype, _, _) = struct.unpack('>HHH', fp.read(6))
            if fmttype == 0:
                char2gid.update(enumerate(struct.unpack('>256B', fp.read(256))))
            elif fmttype == 2:
                subheaderkeys = struct.unpack('>256H', fp.read(512))
                firstbytes = [0]*8192
                for (i, k) in enumerate(subheaderkeys):
                    firstbytes[k//8] = i
                nhdrs = max(subheaderkeys)//8 + 1
                hdrs = []
                for i in range(nhdrs):
                    (firstcode, entcount, delta, offset) = struct.unpack('>HHhH', fp.read(8))
                    hdrs.append((i, firstcode, entcount, delta, fp.tell()-2+offset))
                for (i, firstcode, entcount, delta, pos) in hdrs:
                    if not entcount:
                        continue
                    first = firstcode + (firstbytes[i] << 8)
                    fp.seek(pos)
                    for c in range(entcount):
                        gid = struct.unpack('>H', fp.read(2))
                        if gid:
                            gid += delta
                        char2gid[first+c] = gid
            elif fmttype == 4:
                (segcount, _1, _2, _3) = struct.unpack('>HHHH', fp.read(8))
                segcount //= 2
                ecs = struct.unpack('>%dH' % segcount, fp.read(2*segcount))
                fp.read(2)
                scs = struct.unpack('>%dH' % segcount, fp.read(2*segcount))
                idds = struct.unpack('>%dh' % segcount, fp.read(2*segcount))
                pos = fp.tell()
                idrs = struct.unpack('>%dH' % segcount, fp.read(2*segcount))
                for (ec, sc, idd, idr) in zip(ecs, scs, idds, idrs):
                    if idr:
                        fp.seek(pos+idr)
                        for c in range(sc, ec+1):
                            char2gid[c] = (struct.unpack('>H', fp.read(2))[0] + idd) & 0xffff
                    else:
                        for c in range(sc, ec+1):
                            char2gid[c] = (c + idd) & 0xffff
            else:
                assert False, str(('Unhandled', fmttype))
        # create unicode map
        unicode_map = FileUnicodeMap()
        for (char, gid) in char2gid.items():
            unicode_map.add_cid2unichr(gid, char)
        return unicode_map




from typing import List

from ..PDFGraphicState import PDFGraphicState, PDFColorSpace
from ..Curves import CurvePath

from ._base import PDFItem

class PDFShape(PDFItem):
    def __init__(self, gstate: PDFGraphicState, evenodd: bool, paths: List[CurvePath]):
        self.gstate = gstate
        self.evenodd = evenodd
        self.paths = paths


from ...utils import settings
from ..PSStackParser import LIT

LITERAL_CRYPT = LIT('Crypt')

# Abbreviation of Filter names in PDF 4.8.6. "Inline Images"
LITERALS_FLATE_DECODE = (LIT('FlateDecode'), LIT('Fl'))
LITERALS_LZW_DECODE = (LIT('LZWDecode'), LIT('LZW'))
LITERALS_ASCII85_DECODE = (LIT('ASCII85Decode'), LIT('A85'))
LITERALS_ASCIIHEX_DECODE = (LIT('ASCIIHexDecode'), LIT('AHx'))
LITERALS_RUNLENGTH_DECODE = (LIT('RunLengthDecode'), LIT('RL'))
LITERALS_CCITTFAX_DECODE = (LIT('CCITTFaxDecode'), LIT('CCF'))
LITERALS_DCT_DECODE = (LIT('DCTDecode'), LIT('DCT'))


from typing import List

from pdfmajor.utils import MATRIX_IDENTITY, mult_matrix
from pdfmajor.parser.PSStackParser import literal_name
from pdfmajor.parser.PDFStream import PDFStream, list_value, dict_value
from pdfmajor.parser.constants import LITERAL_FORM, LITERAL_IMAGE
from .state import PDFStateStack, PDFColorSpace
from .state.Curves import CurveMethod, CurvePath, CurvePoint

class PDFCommands:
    commands = {}
    class InvalidOperation(Exception): pass
    class RepeatedCommand(Exception): pass

    @classmethod
    def __add(cls, cmd_name: str, cmd: callable):
        if not (cmd_name in cls.commands.keys()):
            cls.commands[cmd_name] = cmd
            return cmd
        raise cls.RepeatedCommand(cmd_name)

    @classmethod
    def add(cls, *cmd_names: List[str]):
        def wrap(cmd: callable):
            for cmd_name in cmd_names:
                cls.__add(cmd_name, cmd)
            return cmd
        return wrap





import struct
import chardet  

# from sys import maxint as INF #doesn't work anymore under Python3,
# but PDF still uses 32 bits ints
int2byte = struct.Struct(">B").pack

def make_compat_bytes(in_str):
    "In Py2, does nothing. In Py3, converts to bytes, encoding to unicode."
    assert isinstance(in_str, str), str(type(in_str))
    return in_str.encode()

def make_compat_str(in_str):
    "In Py2, does nothing. In Py3, converts to string, guessing encoding."
    assert isinstance(in_str, (bytes, str, unicode)), str(type(in_str))
    if isinstance(in_str, bytes):
        enc = chardet.detect(in_str)
        in_str = in_str.decode(enc['encoding'])
    return in_str

def compatible_encode_method(bytesorstring, encoding='utf-8', erraction='ignore'):
    if isinstance(bytesorstring, str): return bytesorstring
    assert isinstance(bytesorstring, bytes), str(type(bytesorstring))
    return bytesorstring.decode(encoding, erraction)

##  PNG Predictor
##
def apply_png_predictor(pred, colors, columns, bitspercomponent, data):
    if bitspercomponent != 8:
        # unsupported
        raise ValueError("Unsupported `bitspercomponent': %d" %
                         bitspercomponent)
    nbytes = colors * columns * bitspercomponent // 8
    i = 0
    buf = b''
    line0 = b'\x00' * columns
    for i in range(0, len(data), nbytes+1):
        ft = data[i]
        i += 1
        line1 = data[i:i+nbytes]
        line2 = b''
        if ft == 0:
            # PNG none
            line2 += line1
        elif ft == 1:
            # PNG sub (UNTESTED)
            c = 0
            for b in line1:
                c = (c+b) & 255
                line2 += int2byte(c)
        elif ft == 2:
            # PNG up
            for (a, b) in zip(line0, line1):
                c = (a+b) & 255
                line2 += int2byte(c)
        elif ft == 3:
            # PNG average (UNTESTED)
            c = 0
            for (a, b) in zip(line0, line1):
                c = ((c+a+b)//2) & 255
                line2 += int2byte(c)
        else:
            # unsupported
            raise ValueError("Unsupported predictor value: %d" % ft)
        buf += line2
        line0 = line2
    return buf




##  Utility functions
##

# isnumber
def isnumber(x):
    return isinstance(x, (int, float))

# uniq
def uniq(objs):
    """Eliminates duplicated elements."""
    done = set()
    for obj in objs:
        if obj in done:
            continue
        done.add(obj)
        yield obj
    return


# csort
def csort(objs, key):
    """Order-preserving sorting function."""
    idxs = dict((obj, i) for (i, obj) in enumerate(objs))
    return sorted(objs, key=lambda obj: (key(obj), idxs[obj]))


# drange
def drange(v0, v1, d):
    """Returns a discrete range."""
    assert v0 < v1, str((v0, v1, d))
    return range(int(v0)//d, int(v1+d)//d)

# pick
def pick(seq, func, maxobj=None):
    """Picks the object obj where func(obj) has the highest value."""
    maxscore = None
    for obj in seq:
        score = func(obj)
        if maxscore is None or maxscore < score:
            (maxscore, maxobj) = (score, obj)
    return maxobj


# choplist
def choplist(n, seq):
    """Groups every n elements of the list."""
    r = []
    for x in seq:
        r.append(x)
        if len(r) == n:
            yield tuple(r)
            r = []
    return


# nunpack
def nunpack(s, default=0):
    """Unpacks 1 to 4 or 8 byte integers (big endian)."""
    l = len(s)
    if not l:
        return default
    elif l == 1:
        return ord(s)
    elif l == 2:
        return struct.unpack('>H', s)[0]
    elif l == 3:
        return struct.unpack('>L', b'\x00'+s)[0]
    elif l == 4:
        return struct.unpack('>L', s)[0]
    elif l == 8:
        return struct.unpack('>Q', s)[0]
    else:
        raise TypeError('invalid length: %d' % l)

##  Plane
##
##  A set-like data structure for objects placed on a plane.
##  Can efficiently find objects in a certain rectangular area.
##  It maintains two parallel lists of objects, each of
##  which is sorted by its x or y coordinate.
##
class Plane(object):

    def __init__(self, bbox, gridsize=50):
        self._seq = []          # preserve the object order.
        self._objs = set()
        self._grid = {}
        self.gridsize = gridsize
        (self.x0, self.y0, self.x1, self.y1) = bbox
        return

    def __repr__(self):
        return ('<Plane objs=%r>' % list(self))

    def __iter__(self):
        return ( obj for obj in self._seq if obj in self._objs )

    def __len__(self):
        return len(self._objs)

    def __contains__(self, obj):
        return obj in self._objs

    def _getrange(self, bbox):
        (x0, y0, x1, y1) = bbox
        if (x1 <= self.x0 or self.x1 <= x0 or
            y1 <= self.y0 or self.y1 <= y0): return
        x0 = max(self.x0, x0)
        y0 = max(self.y0, y0)
        x1 = min(self.x1, x1)
        y1 = min(self.y1, y1)
        for y in drange(y0, y1, self.gridsize):
            for x in drange(x0, x1, self.gridsize):
                yield (x, y)
        return

    # extend(objs)
    def extend(self, objs):
        for obj in objs:
            self.add(obj)
        return

    # add(obj): place an object.
    def add(self, obj):
        for k in self._getrange((obj.x0, obj.y0, obj.x1, obj.y1)):
            if k not in self._grid:
                r = []
                self._grid[k] = r
            else:
                r = self._grid[k]
            r.append(obj)
        self._seq.append(obj)
        self._objs.add(obj)
        return

    # remove(obj): displace an object.
    def remove(self, obj):
        for k in self._getrange((obj.x0, obj.y0, obj.x1, obj.y1)):
            try:
                self._grid[k].remove(obj)
            except (KeyError, ValueError):
                pass
        self._objs.remove(obj)
        return

    # find(): finds objects that are in a certain area.
    def find(self, bbox):
        (x0, y0, x1, y1) = bbox
        done = set()
        for k in self._getrange(bbox):
            if k not in self._grid:
                continue
            for obj in self._grid[k]:
                if obj in done:
                    continue
                done.add(obj)
                if (obj.x1 <= x0 or x1 <= obj.x0 or
                    obj.y1 <= y0 or y1 <= obj.y0):
                    continue
                yield obj
        return







from typing import List
from .PDFColorSpace import PDFColorSpace

class PDFColor:
    def __init__(self, color_space: PDFColorSpace, *values: List[float]):
        self.color_space: PDFColorSpace = color_space
        self.values: list = values

    def copy(self):
        return self.__class__(
            self.color_space,
            *self.values
        )
    
    def __repr__(self):
        return f"<PDFColor space={self.color_space} values={self.values}/>"



from pdfmajor.utils import get_logger, settings
from pdfmajor.parser.PSStackParser import literal_name
from pdfmajor.parser.PDFStream import list_value
from pdfmajor.parser.PDFStream import dict_value
from pdfmajor.parser.PDFStream import resolve1
from pdfmajor.parser.constants import LITERAL_FONT

from .fonts import PDFType1Font, PDFTrueTypeFont, PDFType3Font, PDFCIDFont, PDFFont

log = get_logger(__file__)

def get_font(objid: int, spec: dict, cached_fonts: dict = {}):
    if objid and objid in cached_fonts:
        font = cached_fonts[objid]
    else:
        log.debug('get_font: create: objid=%r, spec=%r', objid, spec)
        if settings.STRICT:
            if spec['Type'] is not LITERAL_FONT:
                raise PDFFont.PDFFontError('Type is not /Font')
        # Create a Font object.
        if 'Subtype' in spec:
            subtype = literal_name(spec['Subtype'])
        else:
            if settings.STRICT:
                raise PDFFont.PDFFontError('Font Subtype is not specified.')
            subtype = 'Type1'
        if subtype in ('Type1', 'MMType1'):
            # Type1 Font
            font = PDFType1Font(spec)
        elif subtype == 'TrueType':
            # TrueType Font
            font = PDFTrueTypeFont(spec)
        elif subtype == 'Type3':
            # Type3 Font
            font = PDFType3Font(spec)
        elif subtype in ('CIDFontType0', 'CIDFontType2'):
            # CID Font
            font = PDFCIDFont(spec)
        elif subtype == 'Type0':
            # Type0 Font
            dfonts = list_value(spec['DescendantFonts'])
            assert dfonts
            subspec = dict_value(dfonts[0]).copy()
            for k in ('Encoding', 'ToUnicode'):
                if k in spec:
                    subspec[k] = resolve1(spec[k])
            font = get_font(None, subspec, cached_fonts)
        else:
            if settings.STRICT:
                raise PDFFont.PDFFontError('Invalid Font spec: %r' % spec)
            font = PDFType1Font(spec)  # this is so wrong!
        if objid:
            cached_fonts[objid] = font
    return font






from pdfmajor.utils import mult_matrix, MATRIX_IDENTITY, apply_matrix_pt, get_logger
from pdfmajor.parser.PSStackParser import literal_name
from pdfmajor.parser.PDFStream import PDFStream, list_value, dict_value
from pdfmajor.parser.constants import LITERAL_FORM, LITERAL_IMAGE

from .PDFCommands import PDFCommands
from .state import PDFStateStack, PDFGraphicState, LTTextBlock
from .state.Curves import CurveMethod, CurvePath, CurvePoint
# from .state.PDFItem import PDFImage, PDFShape, PDFText, PDFXObject
from .state import make_char_block, make_curve, make_image, make_xobject

log = get_logger('commands')

# No Support
# @PDFCommands.add('W','W_a', 'sh', 'ET', 'BX', 'EX' ,'BI', 'ID', 'gs')
# def do_no_support(stack: PDFStateStack) -> PDFStateStack:
#     # log.warning("Unsupported command")
#     return stack

# marked content operators
@PDFCommands.add('MP')
def do_MP(stack: PDFStateStack, tag) -> PDFStateStack:
    log.warning(f"Unsupported command - do_tag {[tag]}")
    # self.device.do_tag(tag)
    return stack

@PDFCommands.add('DP')
def do_DP(stack: PDFStateStack, tag, props=None) -> PDFStateStack:
    log.warning(f"Unsupported command - do_tag {[tag.name, type(tag), props]}")
    # self.device.do_tag(tag)
    return stack

@PDFCommands.add('BMC')
def do_BMC(stack: PDFStateStack, tag) -> PDFStateStack:
    log.warning(f"Unsupported command - begin_tag {[tag.name, type(tag)]}")
    # self.device.begin_tag(tag)
    return stack

@PDFCommands.add('BDC')
def do_BDC(stack: PDFStateStack, tag, props=None) -> PDFStateStack:
    log.warning(f"Unsupported command - begin_tag {[tag.name, type(tag), props]}")
    # self.device.begin_tag(tag)
    return stack

@PDFCommands.add('EMC')
def do_EMC(stack: PDFStateStack) -> PDFStateStack:
    # self.device.end_tag(tag)
    log.warning("Unsupported command - end_tag")
    return stack

@PDFCommands.add('q')
def do_q(stack: PDFStateStack) -> PDFStateStack:
    stack.gstack.append([
        stack.t_matrix,
        stack.text.copy(),
        stack.graphics.copy()
    ])
    return stack

@PDFCommands.add('Q')
def do_Q(stack: PDFStateStack) -> PDFStateStack:
    if len(stack.gstack) > 0:
        stack.t_matrix, stack.text, stack.graphicstate = stack.gstack.pop()
    return stack

# concat-matrix
@PDFCommands.add('cm')
def do_cm(stack: PDFStateStack, a1, b1, c1, d1, e1, f1) -> PDFStateStack:
    stack.t_matrix = mult_matrix((a1, b1, c1, d1, e1, f1), stack.t_matrix)
    return stack

# setlinewidth
@PDFCommands.add('w')
def do_w(stack: PDFStateStack, linewidth) -> PDFStateStack:
    stack.graphics.linewidth = linewidth
    return stack

# setlinecap
@PDFCommands.add('J')
def do_J(stack: PDFStateStack, linecap) -> PDFStateStack:
    stack.graphics.linecap = linecap
    return stack

# setlinejoin
@PDFCommands.add('j')
def do_j(stack: PDFStateStack, linejoin) -> PDFStateStack:
    stack.graphics.linejoin = linejoin
    return stack

# setmiterlimit
@PDFCommands.add('M')
def do_M(stack: PDFStateStack, miterlimit) -> PDFStateStack:
    stack.graphics.miterlimit = miterlimit
    return stack

# setdash
@PDFCommands.add('d')
def do_d(stack: PDFStateStack, dash, phase) -> PDFStateStack:
    stack.graphics.dash = (dash, phase)
    return stack

# setintent
@PDFCommands.add('ri')
def do_ri(stack: PDFStateStack, intent) -> PDFStateStack:
    stack.graphics.intent = intent
    return stack

# setflatness
@PDFCommands.add('i')
def do_i(stack: PDFStateStack, flatness) -> PDFStateStack:
    stack.graphics.flatness = flatness
    return stack

# moveto
@PDFCommands.add('m')
def do_m(stack: PDFStateStack, x: float, y: float) -> PDFStateStack:
    stack.curvestacks.append(CurvePath(
        CurvePath.METHOD.MOVE_TO, 
        CurvePoint(x, y)
    ))
    return stack

# lineto
@PDFCommands.add('l')
def do_l(stack: PDFStateStack, x: float, y: float) -> PDFStateStack:
    stack.curvestacks.append(CurvePath(
        CurvePath.METHOD.LINE_TO, 
        CurvePoint(x, y)
    ))
    return stack

# curveto
@PDFCommands.add('c')
def do_c(stack: PDFStateStack, 
        x1: float, y1: float, 
        x2: float, y2: float, 
        x3: float, y3: float ) -> PDFStateStack:
    stack.curvestacks.append(CurvePath(
        CurvePath.METHOD.CURVE_BOTH_TO, 
        CurvePoint(x1, y1), 
        CurvePoint(x2, y2), 
        CurvePoint(x3, y3),
    ))
    return stack

# urveto
@PDFCommands.add('v')
def do_v(stack: PDFStateStack, 
        x2: float, y2: float, 
        x3: float, y3: float) -> PDFStateStack:
    stack.curvestacks.append(CurvePath(
        CurvePath.METHOD.CURVE_NEXT_TO, 
        CurvePoint(x2, y2), 
        CurvePoint(x3, y3),
    ))
    return stack

# rveto
@PDFCommands.add('y')
def do_y(stack: PDFStateStack, 
        x1: float, y1: float, 
        x3: float, y3: float) -> PDFStateStack:
    stack.curvestacks.append(CurvePath(
        CurvePath.METHOD.CURVE_FIRST_TO, 
        CurvePoint(x1, y1), 
        CurvePoint(x3, y3)
    ))
    return stack

# closepath
@PDFCommands.add('h')
def do_h(stack: PDFStateStack) -> PDFStateStack:
    stack.curvestacks.append(CurvePath(CurvePath.METHOD.CLOSE_PATH))
    return stack

# rectangle
@PDFCommands.add('re')
def do_re(stack: PDFStateStack, x, y, w, h) -> PDFStateStack:
    stack.curvestacks.extend([
        CurvePath(CurvePath.METHOD.MOVE_TO, CurvePoint(x, y) ),
        CurvePath(CurvePath.METHOD.LINE_TO, CurvePoint(x+w, y) ),
        CurvePath(CurvePath.METHOD.LINE_TO, CurvePoint(x+w, y+h) ),
        CurvePath(CurvePath.METHOD.LINE_TO, CurvePoint(x, y+h) ),
        CurvePath(CurvePath.METHOD.CLOSE_PATH)
    ])
    return stack

# stroke, fill
@PDFCommands.add('S', 'f', 'F', 'B')
def do_complete_path(stack: PDFStateStack) -> PDFStateStack:
    stack.complete_layout_items.append(make_curve(
        stack.t_matrix,
        stack.graphics.copy(),
        False,
        stack.curvestacks
    ))
    stack.curvestacks = []
    return stack

# sroke, fill-even-odd
@PDFCommands.add('f_a', 'B_a')
def do_complete_path_evenodd(stack: PDFStateStack) -> PDFStateStack:
    stack.complete_layout_items.append(make_curve(
        stack.t_matrix,
        stack.graphics.copy(),
        True,
        stack.curvestacks
    ))
    stack.curvestacks = []
    return stack

# close-and-stroke (do_h, do_S)
@PDFCommands.add('s', 'b')
def do_close_complete(stack: PDFStateStack) -> PDFStateStack:
    stack.curvestacks.append(CurvePath(
        CurvePath.METHOD.CLOSE_PATH
    ))
    stack.complete_layout_items.append(make_curve(
        stack.t_matrix,
        stack.graphics.copy(),
        False,
        stack.curvestacks
    ))
    stack.curvestacks = []
    return stack

# close-fill-and-stroke-even-odd
@PDFCommands.add('b_a')
def do_close_complete_evenodd(stack: PDFStateStack) -> PDFStateStack:
    stack.curvestacks.append(CurvePath(
        CurvePath.METHOD.CLOSE_PATH
    ))
    stack.complete_layout_items.append(make_curve(
        stack.t_matrix,
        stack.graphics.copy(),
        True,
        stack.curvestacks
    ))
    stack.curvestacks = []
    return stack

# close-only (empty path)
@PDFCommands.add('n')
def do_n(stack: PDFStateStack) -> PDFStateStack:
    stack.curvestacks = []
    return stack


# setcolorspace-stroking
@PDFCommands.add('CS')
def do_CS(stack: PDFStateStack, name: str) -> PDFStateStack:
    try:
        stack.graphics.scolspace = stack.colorspace_map[literal_name(name)]
        return stack
    except KeyError:
        raise PDFCommands.InvalidOperation('Undefined ColorSpace: %r' % name)

# setcolorspace-non-strokine
@PDFCommands.add('cs')
def do_cs(stack: PDFStateStack, name: str) -> PDFStateStack:
    try:
        stack.graphics.ncolspace = stack.colorspace_map[literal_name(name)]
        return stack
    except KeyError:
        raise PDFCommands.InvalidOperation('Undefined ColorSpace: %r' % name)

# setgray-stroking
@PDFCommands.add('G')
def do_G(stack: PDFStateStack, gray: float) -> PDFStateStack:
    stack.graphics.set_stroke_color(
        stack.colorspace_map['DeviceGray'], 
        gray
    )
    return stack

# setgray-non-stroking
@PDFCommands.add('g')
def do_g(stack: PDFStateStack, gray: float) -> PDFStateStack:
    stack.graphics.set_nostroke_color(
        stack.colorspace_map['DeviceGray'], 
        gray
    )
    return stack


# setrgb-stroking
@PDFCommands.add('RG')
def do_RG(stack: PDFStateStack, r, g, b) -> PDFStateStack:
    stack.graphics.set_stroke_color(
        stack.colorspace_map['DeviceRGB'], 
        round(r*255), 
        round(g*255), 
        round(b*255)
    )
    return stack

# setrgb-non-stroking
@PDFCommands.add('rg')
def do_rg(stack: PDFStateStack, r, g, b) -> PDFStateStack:
    stack.graphics.set_nostroke_color(
        stack.colorspace_map['DeviceRGB'], 
        round(r*255), 
        round(g*255), 
        round(b*255)
    )
    return stack

# setcmyk-stroking
@PDFCommands.add('K')
def do_K(stack: PDFStateStack, c, m, y, k) -> PDFStateStack:
    stack.graphics.set_stroke_color(
        stack.colorspace_map['DeviceCMYK'], 
        c, m, y, k
    )
    return stack

# setcmyk-non-stroking
@PDFCommands.add('k')
def do_k(stack: PDFStateStack, c, m, y, k) -> PDFStateStack:
    stack.graphics.set_stroke_color(
        stack.colorspace_map['DeviceCMYK'], 
        c, m, y, k
    )
    return stack

@PDFCommands.add('SCN','SC')
def do_SCN(stack: PDFStateStack) -> PDFStateStack:
    if stack.graphics.scolspace is None:
        raise PDFCommands.InvalidOperation('No colorspace specified!')
    components = stack.pop(stack.graphics.scolspace.ncomponents)
    stack.graphics.set_stroke_color(
        stack.graphics.scolspace,
        *components
    )
    return stack

@PDFCommands.add('scn','sc')
def do_scn(stack: PDFStateStack) -> PDFStateStack:
    if stack.graphics.ncolspace is None:
        raise PDFCommands.InvalidOperation('No colorspace specified!')
    components = stack.pop(stack.graphics.ncolspace.ncomponents)
    stack.graphics.set_nostroke_color(
        stack.graphics.ncolspace,
        *components
    )
    return stack

# begin-text
@PDFCommands.add('BT')
def do_BT(stack: PDFStateStack) -> PDFStateStack:
    stack.text.matrix = MATRIX_IDENTITY
    stack.text.linematrix = [0, 0]
    if stack.current_textblock is None:
        stack.current_textblock = LTTextBlock()
    else:
        raise PDFCommands.InvalidOperation("Last TextBlock did not clear")
    return stack

@PDFCommands.add('ET')
def do_ET(stack: PDFStateStack) -> PDFStateStack:
    stack.complete_layout_items.append(stack.current_textblock)
    stack.current_textblock = None
    return stack

# setcharspace
@PDFCommands.add('Tc')
def do_Tc(stack: PDFStateStack, charspace) -> PDFStateStack:
    stack.text.charspace = charspace
    return stack

# setwordspace
@PDFCommands.add('Tw')
def do_Tw(stack: PDFStateStack, wordspace) -> PDFStateStack:
    stack.text.wordspace = wordspace
    return stack

# textscale
@PDFCommands.add('Tz')
def do_Tz(stack: PDFStateStack, scaling) -> PDFStateStack:
    stack.text.scaling = scaling
    return stack

# setleading
@PDFCommands.add('TL')
def do_TL(stack: PDFStateStack, leading) -> PDFStateStack:
    stack.text.leading = -leading
    return stack

# selectfont
@PDFCommands.add('Tf')
def do_Tf(stack: PDFStateStack, fontid, fontsize) -> PDFStateStack:
    stack.text.fontsize = fontsize
    try:
        stack.text.font = stack.fontmap[literal_name(fontid)]
        return stack
    except KeyError:
        raise PDFCommands.InvalidOperation('Undefined Font id: %r' % fontid)

# setrendering
@PDFCommands.add('Tr')
def do_Tr(stack: PDFStateStack, render) -> PDFStateStack:
    stack.text.render = render
    return stack

# settextrise
@PDFCommands.add('Ts')
def do_Ts(stack: PDFStateStack, rise) -> PDFStateStack:
    stack.text.rise = rise
    return stack

# text-move
@PDFCommands.add('Td')
def do_Td(stack: PDFStateStack, tx, ty) -> PDFStateStack:
    (a, b, c, d, e, f) = stack.text.matrix
    stack.text.matrix = (a, b, c, d, tx*a+ty*c+e, tx*b+ty*d+f)
    stack.text.linematrix = [0, 0]
    return stack

# text-move
@PDFCommands.add('TD')
def do_TD(stack: PDFStateStack, tx, ty) -> PDFStateStack:
    (a, b, c, d, e, f) = stack.text.matrix
    stack.text.matrix = (a, b, c, d, tx*a+ty*c+e, tx*b+ty*d+f)
    stack.text.leading = ty
    stack.text.linematrix = [0, 0]
    return stack

# textmatrix
@PDFCommands.add('Tm')
def do_Tm(stack: PDFStateStack, a, b, c, d, e, f) -> PDFStateStack:
    stack.text.matrix = (a, b, c, d, e, f)
    stack.text.linematrix = [0, 0]
    return stack

# nextline
@PDFCommands.add('T_a')
def do_T_a(stack: PDFStateStack) -> PDFStateStack:
    (a, b, c, d, e, f) = stack.text.matrix
    stack.text.matrix = (a, b, c, d, stack.text.leading*c+e, stack.text.leading*d+f)
    stack.text.linematrix = [0, 0]
    return stack

# show-pos
@PDFCommands.add('TJ')
def do_TJ(stack: PDFStateStack, seq: bytearray) -> PDFStateStack:
    if stack.text.font is None:
        raise PDFCommands.InvalidOperation("No Font Specified")
    if stack.current_textblock is None:
        raise PDFCommands.InvalidOperation("No TextBlock initilized")
    stack.current_textblock.add_char_block(
        seq,
        stack.t_matrix,
        stack.text.copy(),
        stack.graphics.ncolor.copy()
    )
    return stack

# show
@PDFCommands.add('Tj')
def do_Tj(stack: PDFStateStack, b: bytes) -> PDFStateStack:
    return do_TJ(stack, [b])

# quote
@PDFCommands.add('_q')
def do__q(stack: PDFStateStack, b: bytes) -> PDFStateStack:
    return do_TJ(do_T_a(stack), [b])

# doublequote
@PDFCommands.add('_w')
def do__w(stack: PDFStateStack, wordspace, charspace, b) -> PDFStateStack:
    stack.text.wordspace = wordspace
    stack.text.charspace = charspace
    return do_TJ(stack, [b])

@PDFCommands.add('EI')
def do_EI(stack: PDFStateStack, obj) -> PDFStateStack:
    if 'W' in obj and 'H' in obj:
        stack.complete_layout_items.append(make_image(
            obj,  stack.ctm
        ))
    return stack

# invoke an XObject
@PDFCommands.add('Do')
def do_Do(stack: PDFStateStack, xobjid) -> PDFStateStack:
    xobjid = literal_name(xobjid)
    try:
        xobj = PDFStream.validated_stream(stack.xobjmap[xobjid])
    except KeyError:
        raise PDFCommands.InvalidOperation('Undefined xobject id: %r' % xobjid)
    # log.info('Processing xobj: %r', xobj)
    subtype = xobj.get('Subtype')
    if subtype is LITERAL_FORM and 'BBox' in xobj:
        bbox = list_value(xobj['BBox'])
        matrix = list_value(xobj.get('Matrix', MATRIX_IDENTITY))
        # According to PDF reference 1.7 section 4.9.1, XObjects in
        # earlier PDFs (prior to v1.2) use the page's Resources entry
        # instead of having their own Resources entry.
        xobjres = xobj.get('Resources')
        resources = dict_value(xobjres) if xobjres else stack.resources.copy()
        stack.complete_layout_items.append(make_xobject(
            obj=xobj, 
            bbox=bbox, 
            ctm=stack.t_matrix, 
            matrix=matrix, 
            resources=resources
        ))
    elif subtype is LITERAL_IMAGE and 'Width' in xobj and 'Height' in xobj:
        stack.complete_layout_items.append(make_image(
            xobj,  stack.t_matrix
        ))
    else:
        # unsupported xobject type.
        pass
    return stack




from typing import List
from pdfmajor.utils import MATRIX_IDENTITY

from .PDFGraphicState import PDFGraphicState
from .PDFGraphicState.PDFColor import PDFColor
from .PDFGraphicState.PDFColorSpace import PREDEFINED_COLORSPACE, OrderedDict, PDFColorSpace

from .PDFTextState import PDFTextState, PDFFont, get_font

from .Curves import CurvePath

from .layout import LTItem, LTComponent, LTContainer
from .layout import LTTextBlock, LTCharBlock, LTChar
from .layout import LTCurve, LTLine, LTHorizontalLine, LTVerticalLine, LTRect
from .layout import LTXObject
from .layout import LTImage

from .layout import make_char_block, make_curve, make_image, make_xobject

class PDFStateStack:
    def __init__(self):
        self.t_matrix = MATRIX_IDENTITY
        self.text = PDFTextState()
        self.graphics = PDFGraphicState()
        self.gstack : list = []
        self.curvestacks: List[CurvePath] = []
        self.argstack: List[bytes] = []
        self.complete_layout_items : List[LTItem] = []
        self.colorspace_map: OrderedDict = PREDEFINED_COLORSPACE.copy()
        self.fontmap = {}
        self.xobjmap = {}
        self.resources = {}
        self.current_textblock: LTTextBlock = None

    def pop(self, n: int) -> bytearray:
        if n == 0:
            return []
        x = self.argstack[-n:]
        self.argstack = self.argstack[:-n]
        return x


import re
from typing import List
from xml.sax import saxutils

from .writers.xml import XMLMaker
from ..interpreter import PDFInterpreter, PageInterpreter, logging
from ..interpreter import LTImage, LTTextBlock, LTCharBlock, LTChar, LTCurve, LTXObject
from ..interpreter.commands import LTItem
from ..interpreter.commands import LTRect, LTLine
from ..interpreter.commands.state import CurvePath, PDFColor
from ..parser.PSStackParser import PSLiteral
from ..utils import enc

FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')
def convert_capcase_to_snakecase(text: str) -> str:
    s1 = FIRST_CAP_RE.sub(r'\1-\2', text)
    return ALL_CAP_RE.sub(r'\1-\2', s1).lower()

def normalize_color_values(values: list):
    ret = ""
    for i,v in enumerate(values):
        if i > 0:
            ret += ','
        if isinstance(v, PSLiteral):
            ret += enc(str(v.name)).decode()
        else:
            ret += enc(str(v)).decode()
    return ret

def get_color(col: PDFColor, prefix: str = 'color', codec = 'utf-8'):
    if col.color_space is not None:
        return {
            f"{prefix}-type": enc(col.color_space.name).decode(),
            f"{prefix}-values": normalize_color_values(col.values)
        }
    return {}

def convert_to_xml(
    input_file_path: str, 
    output_file_path: str, 
    image_folder_path: str = None,
    dont_export_images: bool = False,
    codec: str = 'utf-8',
    maxpages: int = 0, 
    password: str = None, 
    caching: bool = True, 
    check_extractable: bool = True,
    pagenos: List[int] = None,
    debug_level: int = logging.WARNING,
):
    intepreter = PDFInterpreter(input_file_path, 
        maxpages=maxpages, 
        password=password, 
        caching=caching,
        check_extractable=check_extractable,
        pagenos=pagenos,
        debug_level=debug_level
    )
    with XMLMaker(output_file_path, codec=codec) as xml:
        with xml.elm("pages"):
            for page in intepreter:
                render_page(xml, page)

def render_page(xml: XMLMaker, ltpage: PageInterpreter):
    def render(item: LTItem):
        if isinstance(item, LTCurve):
            place_curve(xml, item)
        elif isinstance(item, LTXObject):
            with xml.elm('xobject', {
                "x0": item.x0,
                "x1": item.x1,
                "y0": item.y0,
                "y1": item.y1,
            }):
                for child in item:
                    render(child)
        elif isinstance(item, LTImage):
            name = item.name
            # if xml.imagewriter is not None:
            #     name = xml.imagewriter.export_image(item)
            xml.singleton('img', attrs={
                'src': enc(name, None),
                "x0": item.x0,
                "x1": item.x1,
                "y0": item.y0,
                "y1": item.y1,
            })
        elif isinstance(item, LTTextBlock):
            place_text_block(xml, item)
    with xml.elm('page', { 
        "id": ltpage.page_num,
        "width": ltpage.width,
        "height": ltpage.height,
    }):
        for child in ltpage:
            render(child)

def place_char(xml: XMLMaker, char: LTChar):
    with xml.elm('char', {
        'x0': char.x0,
        'x1': char.x1,
        'y0': char.y0,
        'y1': char.y1,
    }, no_additional_char=True):
        xml.write(saxutils.escape(char.get_text()), lineend='', deep_space="")

def place_char_block(xml: XMLMaker, char_block: LTCharBlock):
    attr = {
        'size': char_block.size,
        "x0": char_block.x0,
        "x1": char_block.x1,
        "y0": char_block.y0,
        "y1": char_block.y1,
        **get_color(char_block.color, codec=xml.codec),
    }
    for key, value in char_block.font.descriptor.items():
        if key != "Type" and "FontFile" not in key:
            attr[convert_capcase_to_snakecase(key)] = value
    with xml.elm("char-block", attr):
        for char in char_block:
            place_char(xml, char)

def place_text_block(xml: XMLMaker, text_block: LTCharBlock):
    attr = {
        "x0": text_block.x0,
        "x1": text_block.x1,
        "y0": text_block.y0,
        "y1": text_block.y1,
    }
    with xml.elm("text-block", attr):
        for char_block in text_block:
            place_char_block(xml, char_block)

def place_curve(xml: XMLMaker, item: LTCurve):
    attr = {
        "x0": item.x0,
        "x1": item.x1,
        "y0": item.y0,
        "y1": item.y1,
        **get_color(item.stroke, "stroke", codec=xml.codec),
        **get_color(item.fill, "fill", codec=xml.codec),
    }
    if isinstance(item, LTRect):
        return xml.singleton('rect', attr)
    elif isinstance(item, LTLine):
        return xml.singleton('line', attr)

    with xml.elm('curve', attr):
        for path in item.paths:
            with xml.elm('path', { "type": path.method.name }):
                for point in path.points:
                    xml.singleton('point', { 'x': point.x, 'y': point.y })


from io import TextIOWrapper
from contextlib import contextmanager

class HTMLMaker:
    class AccessError(Exception): pass
    def __init__(self, file_path: str, codec: str = 'utf-8'):
        self.file_path = file_path
        self.codec = codec
        self.outfile: TextIOWrapper = None
    
    def __enter__(self):
        self.outfile = open(self.file_path, 'wb')
        self.__levels_deep = 0
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.outfile.close()
        self.outfile = None
    
    def write_raw(self, text):
        if self.outfile is not None:
            if self.codec:
                text = text.encode(self.codec)
            self.outfile.write(text)
        else:
            raise self.AccessError("Attempting to write to file without an open connection")
    
    def write(self, text: str, lineend: str = '\n', deep_space: str = ' '):
        text = deep_space*self.__levels_deep + text + lineend
        self.write_raw(text)

    @contextmanager
    def elm(self, tag_name: str, attrs: dict = None, css: dict = None, nolineend: bool = False, no_additional_char: bool = False, singleton: bool = False):
        lineend = '\n'
        deep_space = ' '
        if no_additional_char:
            lineend = ''
            deep_space = ''
        if attrs is None:
            attrs = {}
        if css is not None:
            attrs['style'] = ";".join([
                f'{name}: {value}'
                for name, value in css.items()
            ])
        attrs = " ".join([
            f'{name}="{value}"'
            for name, value in attrs.items()
        ])
        if not singleton:
            self.write(f"<{tag_name} {attrs}>", lineend=lineend)
            self.__levels_deep += 1
            yield
            self.__levels_deep -= 1
            self.write(f"</{tag_name}>", deep_space=deep_space, lineend='\n' if not nolineend else '')
        else:
            self.write(f"<{tag_name} {attrs} />", lineend=lineend, deep_space=deep_space)
            yield
    
    def singleton(self, tag_name: str, attrs: dict = None, css: dict = None):
        with self.elm(tag_name, attrs, css, singleton=True):
            pass
               